from __future__ import annotations

"""
Contains the base metaclass for providers
"""
import atexit
import asyncio
import contextlib
from types import ModuleType
from typing import Callable, Any, Optional, Coroutine, Type, Union, List, ClassVar, TYPE_CHECKING

from fileio.lib.core import NormalAccessor
from fileio.lib.posix.base import rewrite_async_syntax, func_as_method_coro
from fileio.utils import LazyLib, settings, logger

if TYPE_CHECKING:
    from fileio.providers.tfio import tfFS
    from fsspec.asyn import AsyncFileSystem

    with contextlib.suppress(ImportError):
        from s3transfer.manager import TransferManager

class CloudFileSystemType(type):
    fs: ModuleType = None
    fsa: ModuleType = None
    fs_name: str = None # gcsfs
    boto: ModuleType = None
    s3t: 'TransferManager' = None
    tffs: 'tfFS' = None

    #s3t: 'boto3.s3.transfer.TransferManager' = None

    def is_ready(cls):
        return bool(cls.fsa and cls.fs)
    
    # @classmethod
    def build_gcsfs(cls, **auth_config):
        LazyLib.import_lib('gcsfs')
        import gcsfs
        if auth_config: settings.gcp.update_auth(**auth_config)
        config = settings.gcp.build_gcsfs_config()
        cls.fs = gcsfs.GCSFileSystem(asynchronous = False, **config)
        cls.fsa = rewrite_async_syntax(gcsfs.GCSFileSystem(asynchronous=True, **config), 'gs')
        
        if settings.tfio_enabled:
            from fileio.providers.tfio import tfFS
            if tfFS.enabled:
                logger.info('Leveraging Tensorflow IO support for GCS')
                cls.tffs = tfFS
        
    
    def build_s3fs(cls, **auth_config):
        LazyLib.import_lib('s3fs')
        LazyLib.import_lib('boto3')

        import s3fs
        import boto3
        import boto3.s3.transfer as s3transfer
        from botocore.config import Config as BotoConfig
    
        if auth_config: settings.aws.update_auth(**auth_config)
        config = settings.aws.build_s3fs_config()
        cls.fs = s3fs.S3FileSystem(asynchronous = False, **config)
        cls.fsa = rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **config))
        
        boto_config = BotoConfig(max_pool_connections = settings.core.num_workers * 2)
        cls.boto = boto3.client('s3', region_name = settings.aws.aws_region, config = boto_config)
        transfer_config = s3transfer.TransferConfig(
            use_threads = True,
            max_concurrency = settings.core.num_workers,
        )
        def create_s3t():
            return s3transfer.create_transfer_manager(cls.boto, transfer_config)

        # @property
        # def get_s3t(self):
        #     return create_s3t()
        
        cls.s3t = property(create_s3t)

    
    def build_minio(cls, **auth_config):
        LazyLib.import_lib('s3fs')
        LazyLib.import_lib('boto3')
        
        import s3fs
        import boto3
        import boto3.s3.transfer as s3transfer
        from botocore.config import Config as BotoConfig

        if auth_config: settings.minio.update_auth(**auth_config)
        config = settings.minio.build_s3fs_config()

        cls.fs = s3fs.S3FileSystem(asynchronous=False, **config)
        cls.fsa = rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **config))
                
        boto_config = BotoConfig(
            signature_version = settings.minio.minio_signature_ver,
            max_pool_connections = settings.core.num_workers * 2
        )
        cls.boto = boto3.client(
            's3', 
            region_name = settings.minio.minio_region, 
            endpoint_url = settings.minio.minio_endpoint, 
            aws_access_key_id = settings.minio.minio_access_key, 
            aws_secret_access_key = settings.minio.minio_secret_key, 
            config = boto_config
        )
        transfer_config = s3transfer.TransferConfig(
            use_threads = True,
            max_concurrency = settings.core.num_workers
        )
        def create_s3t():
            return s3transfer.create_transfer_manager(cls.boto, transfer_config)
        
        # cls.s3t = create_s3t
        cls.s3t = property(create_s3t)

    def build_s3c(cls, **auth_config):
        LazyLib.import_lib('s3fs')
        LazyLib.import_lib('boto3')
        
        import s3fs
        import boto3
        import boto3.s3.transfer as s3transfer
        from botocore.config import Config as BotoConfig

        if auth_config: settings.s3_compat.update_auth(**auth_config)
        config = settings.s3_compat.build_s3fs_config()

        cls.fs = s3fs.S3FileSystem(asynchronous=False, **config)
        cls.fsa = rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **config))
                
        boto_config = BotoConfig(
            signature_version = settings.s3_compat.s3_compat_signature_ver,
            max_pool_connections = settings.core.num_workers * 2
        )
        cls.boto = boto3.client(
            's3', 
            region_name = settings.s3_compat.s3_compat_region, 
            endpoint_url = settings.s3_compat.s3_compat_endpoint, 
            aws_access_key_id = settings.s3_compat.s3_compat_access_key, 
            aws_secret_access_key = settings.s3_compat.s3_compat_secret_key, 
            config = boto_config
        )
        transfer_config = s3transfer.TransferConfig(
            use_threads = True,
            max_concurrency = settings.core.num_workers
        )
        def create_s3t():
            return s3transfer.create_transfer_manager(cls.boto, transfer_config)
        
        # cls.s3t = create_s3t
        cls.s3t = property(create_s3t)


    def build_r2(cls, **auth_config):
        LazyLib.import_lib('s3fs')
        LazyLib.import_lib('boto3')
        
        import s3fs
        import boto3
        import boto3.s3.transfer as s3transfer
        from botocore.config import Config as BotoConfig
        from fileio.providers.filesys.cloudflare_r2 import R2FileSystem

        if auth_config: settings.r2.update_auth(**auth_config)
        config = settings.r2.build_s3fs_config()
        cls.fs: R2FileSystem = R2FileSystem(asynchronous=False, **config)
        cls.fsa: R2FileSystem = rewrite_async_syntax(R2FileSystem(asynchronous=True, **config))
        # cls.fs = s3fs.S3FileSystem(asynchronous=False, **config)
        # cls.fsa = rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **config))

                
        boto_config = BotoConfig(
            max_pool_connections = settings.core.num_workers * 2
        )
        cls.boto = boto3.client(
            's3', 
            # region_name = "auto", 
            endpoint_url = settings.r2.r2_endpoint, 
            aws_access_key_id = settings.r2.r2_access_key_id, 
            aws_secret_access_key = settings.r2.r2_secret_access_key, 
            config = boto_config
        )
        transfer_config = s3transfer.TransferConfig(
            use_threads = True,
            max_concurrency = settings.core.num_workers
        )
        def create_s3t():
            return s3transfer.create_transfer_manager(cls.boto, transfer_config)
        
        s3tm = create_s3t()

        # cls.s3t = create_s3t
        cls.s3t = s3tm
        cls.fs.s3tm = s3tm
        cls.fsa.s3tm = s3tm
    

    def build_wasabi(cls, **auth_config):
        LazyLib.import_lib('s3fs')
        LazyLib.import_lib('boto3')
        
        import s3fs
        import boto3
        import boto3.s3.transfer as s3transfer
        from botocore.config import Config as BotoConfig
        from fileio.providers.filesys.wasabi_s3 import WasabiFileSystem

        if auth_config: settings.wasabi.update_auth(**auth_config)
        config = settings.wasabi.build_s3fs_config()
        # cls.fs = s3fs.S3FileSystem(asynchronous=False, **config)
        # cls.fsa = rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **config))
        cls.fs = WasabiFileSystem(asynchronous=False, **config)
        cls.fsa = rewrite_async_syntax(WasabiFileSystem(asynchronous=True, **config))
                
        boto_config = BotoConfig(
            max_pool_connections = settings.core.num_workers * 2
        )
        cls.boto = boto3.client(
            's3', 
            region_name = settings.wasabi.wasabi_region, 
            endpoint_url = settings.wasabi.wasabi_endpoint, 
            aws_access_key_id = settings.wasabi.wasabi_access_key_id, 
            aws_secret_access_key = settings.wasabi.wasabi_secret_access_key, 
            config = boto_config
        )
        transfer_config = s3transfer.TransferConfig(
            use_threads = True,
            max_concurrency = settings.core.num_workers
        )
        def create_s3t():
            return s3transfer.create_transfer_manager(cls.boto, transfer_config)
        
        # cls.s3t = create_s3t
        cls.s3t = property(create_s3t)


    def build_adlfs(cls, **auth_config):
        LazyLib.import_lib('adlfs')
        import adlfs
        if auth_config: settings.azure.update_auth(**auth_config)
        config = settings.azure.build_azurefs_config()
        cls.fs = adlfs.AzureBlobFileSystem(asynchronous = False, **config)
        cls.fsa = rewrite_async_syntax(adlfs.AzureBlobFileSystem(asynchronous=True, **config), 'az')
        

    def build_filesystems(self, force: bool = False, **auth_config):
        """
        Lazily inits the filesystems
        """
        if self.fs is not None and self.fsa is not None and not force: 
            return
        if self.fs_name == 's3fs':
            self.build_s3fs(**auth_config)
        elif self.fs_name == 'minio':
            self.build_minio(**auth_config)
        elif self.fs_name == 'gcsfs':
            self.build_gcsfs(**auth_config)
        elif self.fs_name == 's3c':
            self.build_s3c(**auth_config)
        elif self.fs_name == 'r2':
            self.build_r2(**auth_config)
        elif self.fs_name == 'wasabi':
            self.build_wasabi(**auth_config)
        elif self.fs_name == 'az':
            self.build_adlfs(**auth_config)


    @classmethod
    def reload_filesystem(cls):
        """ 
        Reinitializes the Filesystem
        """
        raise NotImplementedError


def _dummy_func(*args, **kwargs) -> Optional[Any]:
    pass

async def dummy_async_func(*args, **kwargs)  -> Optional[Any]:
    pass

def create_method_fs(cfs: Type[CloudFileSystemType], name: Union[str, List[str]],  func: Optional[Callable] = None, fs_type: str = 'fs') -> Optional[Callable]:
    if not hasattr(cfs, fs_type):
        #print(f'{cfs.__name__} has no {fs_type}')
        return _dummy_func
    fs_module = getattr(cfs, fs_type)
    if not isinstance(name, list): name = [name]
    for n in name:
        if hasattr(fs_module, n):
            #print(f'{cfs.__name__}:{fs_module} has func {fs_type}:{n}')
            return func(getattr(fs_module, n)) if func else getattr(fs_module, n)
    #print(f'{cfs.__name__} has no func {fs_type}:{name}')
    return _dummy_func

def create_async_method_fs(cfs: Type[CloudFileSystemType], name: Union[str, List[str]], func: Optional[Callable] = None, fs_type: str = 'fsa') -> Optional[Union[Callable, Coroutine]]:
    if not hasattr(cfs, fs_type):
        return dummy_async_func
    fs_module = getattr(cfs, fs_type)
    if not isinstance(name, list): name = [name]
    for n in name:
        if hasattr(fs_module, n):
            return func(getattr(fs_module, n)) if func else getattr(fs_module, n)
    return dummy_async_func

def create_staticmethod(cfs: Type[CloudFileSystemType], name: Union[str, List[str]], fs_type: str = 'fs'):
    return create_method_fs(cfs, name = name, func = staticmethod, fs_type = fs_type)


def create_async_coro(cfs: Type[CloudFileSystemType], name: Union[str, List[str]], fs_type: str = 'fs'):
    return create_async_method_fs(cfs, name = name, func = func_as_method_coro, fs_type = fs_type)


class BaseAccessor(NormalAccessor):
    """
    Dummy Accessor class
    """

    
    class CloudFileSystem(metaclass=CloudFileSystemType):
        pass
    
    info: Callable = create_staticmethod(CloudFileSystem, 'info')
    stat: Callable = create_staticmethod(CloudFileSystem, 'stat')
    size: Callable = create_staticmethod(CloudFileSystem, 'size')
    exists: Callable = create_staticmethod(CloudFileSystem, 'exists')
    is_dir: Callable = create_staticmethod(CloudFileSystem, 'isdir')
    is_file: Callable = create_staticmethod(CloudFileSystem, 'isfile')
    copy: Callable = create_staticmethod(CloudFileSystem, 'copy')
    copy_file: Callable = create_staticmethod(CloudFileSystem, 'cp_file')
    get_file: Callable = create_staticmethod(CloudFileSystem, 'get_file')
    put_file: Callable = create_staticmethod(CloudFileSystem, 'put_file')
    metadata: Callable = create_staticmethod(CloudFileSystem, ['metadata', 'info'])
    checksum: Callable = create_staticmethod(CloudFileSystem, 'checksum')

    get: Callable = create_staticmethod(CloudFileSystem, 'get')
    put: Callable = create_staticmethod(CloudFileSystem, 'put')

    open: Callable = create_method_fs(CloudFileSystem, 'open')
    listdir: Callable = create_staticmethod(CloudFileSystem, 'ls')    
    walk: Callable = create_staticmethod(CloudFileSystem, 'walk')

    glob: Callable = create_method_fs(CloudFileSystem, 'glob')
    find: Callable = create_method_fs(CloudFileSystem, 'find')
    touch: Callable = create_method_fs(CloudFileSystem, 'touch')
    cat: Callable = create_method_fs(CloudFileSystem, 'cat')
    cat_file: Callable = create_method_fs(CloudFileSystem, 'cat_file')
    
    
    pipe: Callable = create_method_fs(CloudFileSystem, 'pipe')
    pipe_file: Callable = create_method_fs(CloudFileSystem, 'pipe_file')
    
    mkdir: Callable = create_method_fs(CloudFileSystem, 'mkdir')
    makedirs: Callable = create_method_fs(CloudFileSystem, ['makedirs', 'mkdirs'])
    unlink: Callable = create_method_fs(CloudFileSystem, 'rm_file')
    rmdir: Callable = create_method_fs(CloudFileSystem, 'rmdir')
    rename: Callable = create_method_fs(CloudFileSystem, 'rename')
    replace: Callable = create_method_fs(CloudFileSystem, 'rename')
    remove: Callable = create_method_fs(CloudFileSystem, 'rm')
    rm: Callable = create_staticmethod(CloudFileSystem, 'rm')
    rm_file: Callable = create_staticmethod(CloudFileSystem, 'rm_file')
    
    modified: Callable = create_method_fs(CloudFileSystem, 'modified')
    url: Callable = create_method_fs(CloudFileSystem, 'url')
    ukey: Callable = create_method_fs(CloudFileSystem, 'ukey')
    setxattr: Callable = create_method_fs(CloudFileSystem, 'setxattr')
    invalidate_cache: Callable = create_method_fs(CloudFileSystem, 'invalidate_cache')
    
    filesys: 'AsyncFileSystem' = CloudFileSystem.fs
    async_filesys: 'AsyncFileSystem' = CloudFileSystem.fsa

    boto: ClassVar = CloudFileSystem.boto
    s3t: 'TransferManager' = CloudFileSystem.s3t
    tffs: 'tfFS' = CloudFileSystem.tffs
    
    # Async Methods
    async_stat: Callable = create_async_coro(CloudFileSystem, 'stat')
    async_touch: Callable = create_async_coro(CloudFileSystem, 'touch')
    async_ukey: Callable = create_async_coro(CloudFileSystem, 'ukey')
    async_size: Callable = create_async_coro(CloudFileSystem, 'size')
    async_url: Callable = create_async_coro(CloudFileSystem, 'url')
    async_setxattr: Callable = create_async_coro(CloudFileSystem, 'setxattr')
    async_modified: Callable = create_async_coro(CloudFileSystem, 'modified')
    async_invalidate_cache: Callable = create_async_coro(CloudFileSystem, 'invalidate_cache')
    async_rename: Callable = create_async_coro(CloudFileSystem, 'rename')
    async_replace: Callable = create_async_coro(CloudFileSystem, 'rename')

    async_info: Callable = create_async_method_fs(CloudFileSystem, 'async_info')
    async_exists: Callable = create_async_method_fs(CloudFileSystem, 'async_exists')
    async_glob: Callable = create_async_method_fs(CloudFileSystem, 'async_glob')
    async_find: Callable = create_async_method_fs(CloudFileSystem, 'async_find')
    async_is_dir: Callable = create_async_method_fs(CloudFileSystem, 'async_isdir')
    async_is_file: Callable = create_async_method_fs(CloudFileSystem, 'async_is_file')
    async_copy: Callable = create_async_method_fs(CloudFileSystem, 'async_copy')
    async_copy_file: Callable = create_async_method_fs(CloudFileSystem, 'async_cp_file')

    async_pipe: Callable = create_async_method_fs(CloudFileSystem, 'async_pipe')
    async_pipe_file: Callable = create_async_method_fs(CloudFileSystem, 'async_pipe_file')

    async_get: Callable = create_async_coro(CloudFileSystem, 'async_get')
    async_get_file: Callable = create_async_coro(CloudFileSystem, 'async_get_file')
    
    async_put: Callable = create_async_method_fs(CloudFileSystem, 'async_put')
    async_put_file: Callable = create_async_method_fs(CloudFileSystem, 'async_put_file')
    async_metadata: Callable = create_async_method_fs(CloudFileSystem, 'async_info')
    async_open: Callable = create_async_method_fs(CloudFileSystem, '_open')
    async_mkdir: Callable = create_async_method_fs(CloudFileSystem, 'async_mkdir')
    async_makedirs: Callable = create_async_method_fs(CloudFileSystem, 'async_makedirs')
    async_unlink: Callable = create_async_method_fs(CloudFileSystem, 'async_rm_file')
    async_rmdir: Callable = create_async_method_fs(CloudFileSystem, 'async_rmdir')
    async_remove: Callable = create_async_method_fs(CloudFileSystem, 'async_rm')
    async_rm: Callable = create_async_method_fs(CloudFileSystem, 'async_rm')
    async_rm_file: Callable = create_async_coro(CloudFileSystem, 'async_rm_file')
    async_listdir: Callable = create_async_method_fs(CloudFileSystem, ['async_listdir', 'async_list_objects'])
    async_walk: Callable = create_async_method_fs(CloudFileSystem, 'async_walk')

    @classmethod
    def reload_cfs(cls, **kwargs):
        cls.CloudFileSystem.build_filesystems(**kwargs)
        cls.info: Callable = create_staticmethod(cls.CloudFileSystem, 'info')
        cls.stat: Callable = create_staticmethod(cls.CloudFileSystem, 'stat')
        cls.size: Callable = create_staticmethod(cls.CloudFileSystem, 'size')
        cls.exists: Callable = create_staticmethod(cls.CloudFileSystem, 'exists')
        cls.is_dir: Callable = create_staticmethod(cls.CloudFileSystem, 'isdir')
        cls.is_file: Callable = create_staticmethod(cls.CloudFileSystem, 'isfile')
        cls.copy: Callable = create_staticmethod(cls.CloudFileSystem, 'copy')
        cls.copy_file: Callable = create_staticmethod(cls.CloudFileSystem, 'cp_file')
        cls.get_file: Callable = create_staticmethod(cls.CloudFileSystem, 'get_file')
        cls.put_file: Callable = create_staticmethod(cls.CloudFileSystem, 'put_file')
        cls.metadata: Callable = create_staticmethod(cls.CloudFileSystem, ['metadata', 'info'])

        cls.open: Callable = create_method_fs(cls.CloudFileSystem, 'open')
        cls.listdir: Callable = create_staticmethod(cls.CloudFileSystem, 'ls')    
        cls.walk: Callable = create_staticmethod(cls.CloudFileSystem, 'walk')
        cls.glob: Callable = create_staticmethod(cls.CloudFileSystem, 'glob')
        cls.get: Callable = create_staticmethod(cls.CloudFileSystem, 'get')
        cls.put: Callable = create_staticmethod(cls.CloudFileSystem, 'put')
        
        cls.checksum: Callable = create_method_fs(cls.CloudFileSystem, 'checksum')
        cls.cat: Callable = create_staticmethod(cls.CloudFileSystem, 'cat')
        cls.cat_file: Callable = create_staticmethod(cls.CloudFileSystem, 'cat_file')
        
        cls.pipe: Callable = create_staticmethod(cls.CloudFileSystem, 'pipe')
        cls.pipe_file: Callable = create_staticmethod(cls.CloudFileSystem, 'pipe_file')
    

        cls.find: Callable = create_method_fs(cls.CloudFileSystem, 'find')
        cls.touch: Callable = create_method_fs(cls.CloudFileSystem, 'touch')
        
        
        cls.mkdir: Callable = create_method_fs(cls.CloudFileSystem, 'mkdir')
        cls.makedirs: Callable = create_method_fs(cls.CloudFileSystem, ['makedirs', 'mkdirs'])
        cls.unlink: Callable = create_method_fs(cls.CloudFileSystem, 'rm_file')
        cls.rmdir: Callable = create_method_fs(cls.CloudFileSystem, 'rmdir')
        cls.rename : Callable = create_method_fs(cls.CloudFileSystem, 'rename')
        cls.replace : Callable = create_method_fs(cls.CloudFileSystem, 'rename')
        cls.rm : Callable = create_staticmethod(cls.CloudFileSystem, 'rm')
        cls.rm_file : Callable = create_staticmethod(cls.CloudFileSystem, 'rm_file')
        
        cls.remove : Callable = create_method_fs(cls.CloudFileSystem, 'rm')
        cls.modified: Callable = create_method_fs(cls.CloudFileSystem, 'modified')
        cls.setxattr: Callable = create_method_fs(cls.CloudFileSystem, 'setxattr')
        cls.url: Callable = create_method_fs(cls.CloudFileSystem, 'url')
        cls.ukey: Callable = create_method_fs(cls.CloudFileSystem, 'ukey')
        cls.invalidate_cache: Callable = create_method_fs(cls.CloudFileSystem, 'invalidate_cache')
        
        cls.filesys = cls.CloudFileSystem.fs
        cls.async_filesys = cls.CloudFileSystem.fsa
        cls.boto = cls.CloudFileSystem.boto
        cls.s3t = cls.CloudFileSystem.s3t
        cls.tffs = cls.CloudFileSystem.tffs

        
        # Async Methods
        cls.async_stat: Callable = create_async_coro(cls.CloudFileSystem, 'stat')
        cls.async_touch: Callable = create_async_coro(cls.CloudFileSystem, 'touch')
        cls.async_ukey: Callable = create_async_coro(cls.CloudFileSystem, 'ukey')
        cls.async_size: Callable = create_async_coro(cls.CloudFileSystem, 'size')
        cls.async_url: Callable = create_async_coro(cls.CloudFileSystem, 'url')
        cls.async_setxattr: Callable = create_async_coro(cls.CloudFileSystem, 'setxattr')

        cls.async_modified: Callable = create_async_coro(cls.CloudFileSystem, 'modified')
        cls.async_invalidate_cache: Callable = create_async_coro(cls.CloudFileSystem, 'invalidate_cache')
        cls.async_rename: Callable = create_async_coro(cls.CloudFileSystem, 'rename')
        cls.async_replace: Callable = create_async_coro(cls.CloudFileSystem, 'rename')

        cls.async_info: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_info')
        cls.async_exists: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_exists')

        cls.async_glob: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_glob')
        cls.async_find: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_find')
        
        cls.async_cat: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_cat')
        cls.async_cat_file: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_cat_file')
        
        cls.async_pipe: Callable = create_async_coro(cls.CloudFileSystem, 'async_pipe')
        cls.async_pipe_file: Callable = create_async_coro(cls.CloudFileSystem, 'async_pipe_file')
        
        cls.async_is_dir: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_isdir')
        cls.async_is_file: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_isfile')
        cls.async_copy: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_copy')
        cls.async_copy_file: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_cp_file')
        cls.async_get: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_get')
        cls.async_get_file: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_get_file')
        cls.async_put: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_put')
        cls.async_put_file: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_put_file')
        cls.async_metadata: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_info')
        cls.async_open: Callable = create_async_method_fs(cls.CloudFileSystem, '_open')
        cls.async_mkdir: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_mkdir')
        cls.async_makedirs: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_makedirs')
        cls.async_unlink: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_rm_file')
        cls.async_rm_file: Callable = create_async_coro(cls.CloudFileSystem, 'async_rm_file')
        cls.async_rmdir: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_rmdir')
        cls.async_remove: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_rm')
        cls.async_rm: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_rm')
        cls.async_listdir: Callable = create_async_method_fs(cls.CloudFileSystem, ['async_listdir', 'async_list_objects'])
        cls.async_walk: Callable = create_async_method_fs(cls.CloudFileSystem, 'async_walk')




from __future__ import annotations

import atexit
import contextlib
from typing import Callable, Any, Optional, Union, Dict, Type, TYPE_CHECKING

from fileio.lib.posix.meta import CloudFileSystemType, BaseAccessor

if TYPE_CHECKING:
    from fileio.providers.tfio import tfFS    
    with contextlib.suppress(ImportError):
        import gcsfs
    with contextlib.suppress(ImportError):
        import s3fs
    with contextlib.suppress(ImportError):
        import boto3
    with contextlib.suppress(ImportError):
        import adlfs
    with contextlib.suppress(ImportError):
        from fileio.providers.filesys.cloudflare_r2 import R2FileSystem
    with contextlib.suppress(ImportError):
        from fileio.providers.filesys.wasabi_s3 import WasabiFileSystem
    
    with contextlib.suppress(ImportError):
        from s3transfer.manager import TransferManager


    # with contextlib.suppress(ImportError):
    #     from fileio.providers.hffs.filesys import HfFileSystem


class GCP_CloudFileSystem(metaclass=CloudFileSystemType):
    fs: 'gcsfs.GCSFileSystem' = None
    fsa: 'gcsfs.GCSFileSystem' = None
    fs_name: str = 'gcsfs'
    tffs: 'tfFS' = None # Tensorflow IO

class AWS_CloudFileSystem(metaclass=CloudFileSystemType):
    fs: 's3fs.S3FileSystem' = None
    fsa: 's3fs.S3FileSystem' = None
    fs_name: str = 's3fs'
    boto: 'boto3.session.Session' = None
    s3t: 'TransferManager' = None


class Minio_CloudFileSystem(metaclass=CloudFileSystemType):
    fs: 's3fs.S3FileSystem' = None
    fsa: 's3fs.S3FileSystem' = None
    fs_name: str = 'minio'
    boto: 'boto3.session.Session' = None
    s3t: 'TransferManager' = None

class S3Compat_CloudFileSystem(metaclass=CloudFileSystemType):
    fs: 's3fs.S3FileSystem' = None
    fsa: 's3fs.S3FileSystem' = None
    fs_name: str = 's3c'
    boto: 'boto3.session.Session' = None
    s3t: 'TransferManager' = None

    # async def _aonexit(self, *args, **kwargs):
    #     """
    #     On Exit Function
    #     """
    #     await self.fs.s3.close()
    #     await self.fsa.s3.close()

    # def _onexit(self, *args, **kwargs):
    #     """
    #     On Exit Function
    #     """
    #     sync(
    #         self.fs.loop,
    #         self.fs.s3.close
    #     )
    #     sync(
    #         self.fsa.loop,
    #         self.fsa.s3.close
    #     )
    #     # self.fs.s3.close()
    #     # self.fsa.s3.close()

class R2_CloudFileSystem(metaclass=CloudFileSystemType):
    fs: 'R2FileSystem' = None
    fsa: 'R2FileSystem' = None
    fs_name: str = 'r2'
    boto: 'boto3.session.Session' = None
    s3t: 'TransferManager' = None

class Wasabi_CloudFileSystem(metaclass=CloudFileSystemType):
    # fs: 's3fs.S3FileSystem' = None
    # fsa: 's3fs.S3FileSystem' = None
    fs: 'WasabiFileSystem' = None
    fsa: 'WasabiFileSystem' = None
    fs_name: str = 'wasabi'
    boto: 'boto3.session.Session' = None
    s3t: 'TransferManager' = None

class Azure_CloudFileSystem(metaclass=CloudFileSystemType):
    fs: 'adlfs.AzureBlobFileSystem' = None
    fsa: 'adlfs.AzureBlobFileSystem' = None
    fs_name: str = 'azure'

# class HF_CloudFileSystem(metaclass=CloudFileSystemType):
#     fs: 'HfFileSystem' = None
#     fsa: 'HfFileSystem' = None
#     fs_name: str = 'hf'

class GCP_Accessor(BaseAccessor):
    """
    GCP Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(GCP_CloudFileSystem):
        pass

class AWS_Accessor(BaseAccessor):
    """
    AWS Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(AWS_CloudFileSystem):
        pass

class Minio_Accessor(BaseAccessor):
    """
    S3 Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(Minio_CloudFileSystem):
        pass

class S3Compat_Accessor(BaseAccessor):
    """
    S3 Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(S3Compat_CloudFileSystem):
        pass

# class HF_Accessor(BaseAccessor):
#     """
#     HF Filelike Accessor that inherits from BaseAccessor
#     """
#     class CloudFileSystem(HF_CloudFileSystem):
#         pass

class R2_Accessor(BaseAccessor):
    """
    R2 Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(R2_CloudFileSystem):
        pass


class Wasabi_Accessor(BaseAccessor):
    """
    Wasabi Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(Wasabi_CloudFileSystem):
        pass


class Azure_Accessor(BaseAccessor):
    """
    Azure Filelike Accessor that inherits from BaseAccessor
    """
    class CloudFileSystem(Azure_CloudFileSystem):
        pass

AccessorLike = Union[
    BaseAccessor,
    GCP_Accessor,
    AWS_Accessor,
    Minio_Accessor,
    S3Compat_Accessor,
    # HF_Accessor,
    R2_Accessor,
    Wasabi_Accessor,
    Azure_Accessor,
]

CloudFileSystemLike = Union[
    CloudFileSystemType,
    GCP_CloudFileSystem,
    AWS_CloudFileSystem,
    Minio_CloudFileSystem,
    S3Compat_CloudFileSystem,
    # HF_CloudFileSystem,
    R2_CloudFileSystem,
    Wasabi_CloudFileSystem,
    Azure_CloudFileSystem,
]

class AccessorMeta(type):
    """
    AccessorMeta is a metaclass that is used to register file systems
    """

    ax: Dict[str, Union[AccessorLike, Any]] = {}
    fs_map: Dict[str, Type[CloudFileSystemLike]] = {
        'gs': GCP_CloudFileSystem,
        's3': AWS_CloudFileSystem,
        'minio': Minio_CloudFileSystem,
        's3c': S3Compat_CloudFileSystem,
        # 'hf': HF_CloudFileSystem,
        'r2': R2_CloudFileSystem,
        'wsbi': Wasabi_CloudFileSystem,
        'az': Azure_CloudFileSystem,

    }

    def get_gs_accessor(cls, _reset: Optional[bool] = False, **kwargs) -> GCP_Accessor:
        if not cls.ax.get('gs') or _reset:
            GCP_CloudFileSystem.build_filesystems(**kwargs)
            GCP_Accessor.reload_cfs(**kwargs)
            cls.ax['gs'] = GCP_Accessor()
        return cls.ax['gs']
    
    def get_s3_accessor(cls, _reset: Optional[bool] = False, **kwargs) -> AWS_Accessor:
        if not cls.ax.get('s3') or _reset:
            AWS_CloudFileSystem.build_filesystems(**kwargs)
            AWS_Accessor.reload_cfs(**kwargs)
            cls.ax['s3'] = AWS_Accessor()
            # atexit.register(AWS_CloudFileSystem._onexit)
        return cls.ax['s3']
    
    def get_minio_accessor(cls, _reset: Optional[bool] = False, **kwargs) -> Minio_Accessor:
        if not cls.ax.get('minio') or _reset:
            Minio_CloudFileSystem.build_filesystems(**kwargs)
            Minio_Accessor.reload_cfs(**kwargs)
            cls.ax['minio'] = Minio_Accessor()
        return cls.ax['minio']
    
    def get_s3c_accessor(cls, _reset: Optional[bool] = False, **kwargs) -> S3Compat_Accessor:
        if not cls.ax.get('s3c') or _reset:
            S3Compat_CloudFileSystem.build_filesystems(**kwargs)
            S3Compat_Accessor.reload_cfs(**kwargs)
            cls.ax['s3c'] = S3Compat_Accessor()
        return cls.ax['s3c']

    def get_r2_accessor(cls, _reset: Optional[bool] = False, **kwargs) -> R2_Accessor:
        if not cls.ax.get('r2') or _reset:
            R2_CloudFileSystem.build_filesystems(**kwargs)
            R2_Accessor.reload_cfs(**kwargs)
            cls.ax['r2'] = R2_Accessor()
        return cls.ax['r2']

    def get_wsbi_accessor(cls, _reset: Optional[bool] = False, **kwargs) -> Wasabi_Accessor:
        if not cls.ax.get('wsbi') or _reset:
            Wasabi_CloudFileSystem.build_filesystems(**kwargs)
            Wasabi_Accessor.reload_cfs(**kwargs)
            cls.ax['wsbi'] = Wasabi_Accessor()
        return cls.ax['wsbi']
    
    def get_az_accessor(cls, _reset: Optional[bool] = False, **kwargs) -> Azure_Accessor:
        if not cls.ax.get('az') or _reset:
            Azure_CloudFileSystem.build_filesystems(**kwargs)
            Azure_Accessor.reload_cfs(**kwargs)
            cls.ax['az'] = Azure_Accessor()
        return cls.ax['az']

    def get_accessor(
        cls, 
        name: str, 
        _reset: Optional[bool] = False, 
        **kwargs
    ) -> AccessorLike:
        """
        Returns an accessor for the given file system name
        """
        _ax = getattr(cls, f'get_{name}_accessor', None)
        return _ax(_reset=_reset, **kwargs) if _ax else BaseAccessor
    

    def get_fs(
        cls, 
        name: str, 
        _reset: Optional[bool] = False, 
        **kwargs
    ) -> CloudFileSystemLike:
        """
        Returns a file system for the given file system name
        """
        _fs: CloudFileSystemLike = cls.fs_map.get(name)
        if not _fs: return CloudFileSystemType
        if _reset: _fs.build_filesystems(**kwargs)
        return _fs
        

class FileSysManager(metaclass=AccessorMeta):
    pass




"""
File Conversion Classes
"""
import contextlib
from abc import ABC, abstractmethod
from typing import Optional, Dict, Any, Union, List, Tuple, Iterator, TYPE_CHECKING


# from fileio.types.core import File, FileLike
from fileio.lib.types import File, FileLike
from fileio.types.etc import ParsedFile, PreparedFile
from fileio.utils import logger
from fileio.utils.pooler import ThreadPooler
from fileio.utils.configs import settings
from fileio.utils.ops import (
    async_checksum_file, 
    checksum_file,
    fetch_file_from_url,
    async_fetch_file_from_url
) 

from .static import KNOWN_LIGATURES

_tqdm_avail: bool = False
with contextlib.suppress(ImportError):
    from tqdm.auto import tqdm
    __tqdm_avail = True
    import langdetect

if TYPE_CHECKING:
    with contextlib.suppress(ImportError):
        from starlette.requests import Request
        from starlette.datastructures import UploadFile

class BaseConverter(ABC):
    """
    Base class for implementing file converts to transform input documents to text format for ingestion in Elastic.
    """

    method: Optional[str] = None
    enabled: Optional[bool] = None
    progress_bar: Optional[bool] = (settings.enable_progress_bar and _tqdm_avail)
    raise_errors: Optional[bool] = True

    def __init__(
        self,
        remove_numeric_tables: bool = False,
        valid_languages: Optional[List[str]] = None,
        id_hash_keys: Optional[List[str]] = None,
        progress_bar: Optional[bool] = None,
        raise_errors: Optional[bool] = None,
        **kwargs,
    ):
        """
        :param remove_numeric_tables: This option uses heuristics to remove numeric rows from the tables.
                                      The tabular structures in documents might be noise for the reader model if it
                                      does not have table parsing capability for finding answers. However, tables
                                      may also have long strings that could possible candidate for searching answers.
                                      The rows containing strings are thus retained in this option.
        :param valid_languages: validate languages from a list of languages specified in the ISO 639-1
                                (https://en.wikipedia.org/wiki/ISO_639-1) format.
                                This option can be used to add test for encoding errors. If the extracted text is
                                not one of the valid languages, then it might likely be encoding error resulting
                                in garbled text.
        :param id_hash_keys: Generate the document id from a custom list of strings that refer to the document's
            attributes. If you want to ensure you don't have duplicate documents in your DocumentStore but texts are
            not unique, you can modify the metadata and pass e.g. `"meta"` to this field (e.g. [`"content"`, `"meta"`]).
            In this case the id will be generated by using the content and the defined metadata.
        :param progress_bar: Show a progress bar for the conversion.
        """
        super().__init__()

        self.remove_numeric_tables = remove_numeric_tables
        self.valid_languages = valid_languages
        self.id_hash_keys = id_hash_keys
        if progress_bar is not None: self.progress_bar = progress_bar
        if raise_errors is not None: self.raise_errors = raise_errors

        self._kwargs = kwargs
        self.validate_enabled()
    
    @classmethod
    def set_progress_bar(cls, value: bool):
        cls.progress_bar = value

    @classmethod
    def check_enabled(cls, **kwargs):
        """
        Check if the converter is enabled
        """
        return True

    @classmethod
    def validate_enabled(cls):
        """
        Validate if the converter is enabled
        """
        if cls.enabled is None:
            cls.enabled = cls.check_enabled()
        if not cls.enabled:
            if cls.raise_errors:
                raise NotImplementedError(f"Converter {cls.__name__} is not enabled with {cls.method}")
            logger.warning(f"Converter {cls.__name__} is not enabled with {cls.method}")

    def extract_file_metadata(
        self,
        file_path: 'FileLike',
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Extract file metadata like name, size, etc.
        """
        return {}

    async def async_extract_file_metadata(
        self,
        file_path: 'FileLike',
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Extract file metadata like name, size, etc.
        """
        return await ThreadPooler.run_async(self.extract_file_metadata, file_path, **kwargs)

    def prepare_file(
        self,
        file: 'FileLike',
        **kwargs,
    ) -> 'PreparedFile':
        """
        Prepares a single file for extraction
        """
        # file = File(file)
        file = fetch_file_from_url(file) if isinstance(file, str) and file.startswith('http') \
            else File(file)
        
        if not file.exists():
            if self.raise_errors: raise FileNotFoundError(f"File {file} does not exist")
            logger.warning(f"File {file} does not exist")
            return None
        cksum = checksum_file(file)
        kws = {
            'local_path': None if file.is_cloud else file,
            'remote_path': file if file.is_cloud else None,
            'is_tmp': 'tmp' in file.as_posix(),
            'checksum': cksum,
        }
        return PreparedFile.parse_obj(kws)

    async def async_prepare_file(
        self,
        file: 'FileLike',
        **kwargs,
    ) -> 'PreparedFile':
        """
        Prepares a single file for extraction
        """
        file = await async_fetch_file_from_url(file) if isinstance(file, str) and file.startswith('http') \
            else File(file)

        # file = File(file)        
        if not await file.async_exists():
            if self.raise_errors: raise FileNotFoundError(f"File {file} does not exist")
            logger.warning(f"File {file} does not exist")
            return None
        cksum = await async_checksum_file(file)
        kws = {
            'local_path': None if file.is_cloud else file,
            'remote_path': file if file.is_cloud else None,
            'is_tmp': 'tmp' in file.as_posix(),
            'checksum': cksum,
        }
        return PreparedFile.parse_obj(kws)


    async def async_prepare_files(
        self,
        files: Union['FileLike', List['FileLike']] = None,
        request: Optional['Request'] = None, 
        upload_files: Optional[Union[List['UploadFile'], 'UploadFile']] = None, 
        upload_form_keys: Optional[Union[List[str], str]] = None,
        remote_files: Optional[Union[List[str], str]] = None,
        remote_form_keys: Optional[List[str]] = None,
        **kwargs
    ) -> List['PreparedFile']:
        """
        Prepare files for conversion
        """
        if files is not None:
            if not isinstance(files, list): files = [files]
            files = [
                fetch_file_from_url(f)
                if (isinstance(f, str) and f.startswith('http'))
                else File(f)
                for f in files
            ]
            # files = [File(f) for f in files]
            prepared_files = []
            for file in files:
                if not await file.async_exists():
                    if self.raise_errors: raise FileNotFoundError(f"File {file} does not exist")
                    logger.warning(f"File {file} does not exist")
                    continue

                cksum = await async_checksum_file(file)
                kws = {
                    'local_path': None if file.is_cloud else file,
                    'remote_path': file if file.is_cloud else None,
                    'is_tmp': 'tmp' in file.as_posix(),
                    'checksum': cksum,
                }
                prepared_files.append(PreparedFile.parse_obj(kws))
        else:
            prepared_files = await PreparedFile.async_from_request(
                request=request,
                upload_files=upload_files,
                upload_form_keys=upload_form_keys,
                remote_files=remote_files,
                remote_form_keys=remote_form_keys,
                **kwargs
            )

        if not isinstance(prepared_files, list): prepared_files = [prepared_files]
        return prepared_files
    

    @abstractmethod
    def convert(
        self,
        file: Union['PreparedFile', 'FileLike'],
        metadata: Optional[Dict[str, Any]] = None,
        remove_numeric_tables: Optional[bool] = None,
        valid_languages: Optional[List[str]] = None,
        encoding: Optional[str] = "UTF-8",
        id_hash_keys: Optional[List[str]] = None,
        **kwargs,
    ) -> List[ParsedFile]:
        """
        Convert a file to a dictionary containing the text and any associated meta data.
        File converters may extract file meta like name or size. In addition to it, user
        supplied meta data like author, url, external IDs can be supplied as a dictionary.
        :param file: path of the file to convert
        :param metadata: dictionary of meta data key-value pairs to append in the returned document.
        :param remove_numeric_tables: This option uses heuristics to remove numeric rows from the tables.
                                      The tabular structures in documents might be noise for the reader model if it
                                      does not have table parsing capability for finding answers. However, tables
                                      may also have long strings that could possible candidate for searching answers.
                                      The rows containing strings are thus retained in this option.
        :param valid_languages: validate languages from a list of languages specified in the ISO 639-1
                                (https://en.wikipedia.org/wiki/ISO_639-1) format.
                                This option can be used to add test for encoding errors. If the extracted text is
                                not one of the valid languages, then it might likely be encoding error resulting
                                in garbled text.
        :param encoding: Select the file encoding (default is `UTF-8`)
        :param id_hash_keys: Generate the document id from a custom list of strings that refer to the document's
            attributes. If you want to ensure you don't have duplicate documents in your DocumentStore but texts are
            not unique, you can modify the metadata and pass e.g. `"meta"` to this field (e.g. [`"content"`, `"meta"`]).
            In this case the id will be generated by using the content and the defined metadata.
        """
        pass

    
    async def async_convert(
        self,
        file: Union['PreparedFile', 'FileLike'],
        metadata: Optional[Dict[str, Any]] = None,
        remove_numeric_tables: Optional[bool] = None,
        valid_languages: Optional[List[str]] = None,
        encoding: Optional[str] = "UTF-8",
        id_hash_keys: Optional[List[str]] = None,
        **kwargs,
    ) -> List[ParsedFile]:
        """
        Convert a file to a dictionary containing the text and any associated meta data.
        File converters may extract file meta like name or size. In addition to it, user
        supplied meta data like author, url, external IDs can be supplied as a dictionary.
        :param file_path: path of the file to convert
        :param meta: dictionary of meta data key-value pairs to append in the returned document.
        :param remove_numeric_tables: This option uses heuristics to remove numeric rows from the tables.
                                      The tabular structures in documents might be noise for the reader model if it
                                      does not have table parsing capability for finding answers. However, tables
                                      may also have long strings that could possible candidate for searching answers.
                                      The rows containing strings are thus retained in this option.
        :param valid_languages: validate languages from a list of languages specified in the ISO 639-1
                                (https://en.wikipedia.org/wiki/ISO_639-1) format.
                                This option can be used to add test for encoding errors. If the extracted text is
                                not one of the valid languages, then it might likely be encoding error resulting
                                in garbled text.
        :param encoding: Select the file encoding (default is `UTF-8`)
        :param id_hash_keys: Generate the document id from a custom list of strings that refer to the document's
            attributes. If you want to ensure you don't have duplicate documents in your DocumentStore but texts are
            not unique, you can modify the metadata and pass e.g. `"meta"` to this field (e.g. [`"content"`, `"meta"`]).
            In this case the id will be generated by using the content and the defined metadata.
        """
        return await ThreadPooler.run_async(self.convert, file = file,  encoding = encoding, id_hash_keys = id_hash_keys, metadata = metadata, remove_numeric_tables = remove_numeric_tables, valid_languages = valid_languages, **kwargs)

    def validate_language(self, text: str, valid_languages: Optional[List[str]] = None) -> bool:
        """
        Validate if the language of the text is one of valid languages.
        """
        if valid_languages is None: valid_languages = self.valid_languages
        if not valid_languages: return True
        try: lang = langdetect.detect(text)
        except langdetect.lang_detect_exception.LangDetectException: lang = None
        return lang in valid_languages
    
    def get_iterator(self, files: List['PreparedFile'], metadata: List[Dict[str, Any]]) -> Iterator[Tuple['PreparedFile', Dict[str, Any]]]:
        """
        Get an iterator over the files and metadata.
        """
        if self.progress_bar:
            return tqdm(
                zip(files, metadata), total=len(files), disable = not self.progress_bar, desc="Converting files"
            )
        return zip(files, metadata)
        

    def run(  # type: ignore
        self,
        files: Union['FileLike', 'PreparedFile', List['FileLike'], List['PreparedFile']],
        metadata: Optional[Union[Dict[str, str], List[Optional[Dict[str, str]]]]] = None,
        remove_numeric_tables: Optional[bool] = None,
        known_ligatures: Optional[Dict[str, str]] = None,
        valid_languages: Optional[List[str]] = None,
        encoding: Optional[str] = "UTF-8",
        id_hash_keys: Optional[List[str]] = None,
        extract_metadata: Optional[bool] = None,
        **kwargs,
    ) -> List[ParsedFile]:
        """
        Extract text from a file.
        :param file_paths: Path to the files you want to convert
        :param meta: Optional dictionary with metadata that shall be attached to all resulting documents.
                     Can be any custom keys and values.
        :param remove_numeric_tables: This option uses heuristics to remove numeric rows from the tables.
                                      The tabular structures in documents might be noise for the reader model if it
                                      does not have table parsing capability for finding answers. However, tables
                                      may also have long strings that could possible candidate for searching answers.
                                      The rows containing strings are thus retained in this option.
        :param known_ligatures: Some converters tend to recognize clusters of letters as ligatures, such as "ﬀ" (double f).
                                Such ligatures however make text hard to compare with the content of other files,
                                which are generally ligature free. Therefore we automatically find and replace the most
                                common ligatures with their split counterparts. The default mapping is in
                                `haystack.nodes.file_converter.base.KNOWN_LIGATURES`: it is rather biased towards Latin alphabeths
                                but excludes all ligatures that are known to be used in IPA.
                                If no value is provided, this default is created and used.
                                You can use this parameter to provide your own set of ligatures to clean up from the documents.
        :param valid_languages: validate languages from a list of languages specified in the ISO 639-1
                                (https://en.wikipedia.org/wiki/ISO_639-1) format.
                                This option can be used to add test for encoding errors. If the extracted text is
                                not one of the valid languages, then it might likely be encoding error resulting
                                in garbled text.
        :param encoding: Select the file encoding (default is `UTF-8`)
        :param id_hash_keys: Generate the document id from a custom list of strings that refer to the document's
            attributes. If you want to ensure you don't have duplicate documents in your DocumentStore but texts are
            not unique, you can modify the metadata and pass e.g. `"meta"` to this field (e.g. [`"content"`, `"meta"`]).
            In this case the id will be generated by using the content and the defined metadata.
        """
        if known_ligatures is None: known_ligatures = KNOWN_LIGATURES
        if not isinstance(files, list): files = [files]
        prepared_files = [self.prepare_file(file) for file in files]

        if isinstance(metadata, dict) or metadata is None: metadata = [metadata] * len(prepared_files)
        parsed_files: List[ParsedFile] = []

        for prepared_file, file_meta in self.get_iterator(prepared_files, metadata):
            parsed_files.extend(
                self.convert(
                    file=prepared_file,
                    meta=file_meta,
                    remove_numeric_tables=remove_numeric_tables,
                    valid_languages=valid_languages,
                    encoding=encoding,
                    id_hash_keys=id_hash_keys,
                    extract_metadata=extract_metadata,
                    **kwargs,
                )
            )

        # Cleanup ligatures
        for parsed_file in parsed_files:
            for ligature, letters in known_ligatures.items():
                if parsed_file.content is not None: parsed_file.content = parsed_file.content.replace(ligature, letters)
        return parsed_files
    
    async def async_run(  # type: ignore
        self,
        files: Union['FileLike', 'PreparedFile', List['FileLike'], List['PreparedFile']],
        metadata: Optional[Union[Dict[str, str], List[Optional[Dict[str, str]]]]] = None,
        remove_numeric_tables: Optional[bool] = None,
        known_ligatures: Optional[Dict[str, str]] = None,
        valid_languages: Optional[List[str]] = None,
        encoding: Optional[str] = "UTF-8",
        id_hash_keys: Optional[List[str]] = None,
        extract_metadata: Optional[bool] = None,
        **kwargs,
    ) -> List[ParsedFile]:
        """
        Extract text from a file.
        :param file_paths: Path to the files you want to convert
        :param meta: Optional dictionary with metadata that shall be attached to all resulting documents.
                     Can be any custom keys and values.
        :param remove_numeric_tables: This option uses heuristics to remove numeric rows from the tables.
                                      The tabular structures in documents might be noise for the reader model if it
                                      does not have table parsing capability for finding answers. However, tables
                                      may also have long strings that could possible candidate for searching answers.
                                      The rows containing strings are thus retained in this option.
        :param known_ligatures: Some converters tend to recognize clusters of letters as ligatures, such as "ﬀ" (double f).
                                Such ligatures however make text hard to compare with the content of other files,
                                which are generally ligature free. Therefore we automatically find and replace the most
                                common ligatures with their split counterparts. The default mapping is in
                                `haystack.nodes.file_converter.base.KNOWN_LIGATURES`: it is rather biased towards Latin alphabeths
                                but excludes all ligatures that are known to be used in IPA.
                                If no value is provided, this default is created and used.
                                You can use this parameter to provide your own set of ligatures to clean up from the documents.
        :param valid_languages: validate languages from a list of languages specified in the ISO 639-1
                                (https://en.wikipedia.org/wiki/ISO_639-1) format.
                                This option can be used to add test for encoding errors. If the extracted text is
                                not one of the valid languages, then it might likely be encoding error resulting
                                in garbled text.
        :param encoding: Select the file encoding (default is `UTF-8`)
        :param id_hash_keys: Generate the document id from a custom list of strings that refer to the document's
            attributes. If you want to ensure you don't have duplicate documents in your DocumentStore but texts are
            not unique, you can modify the metadata and pass e.g. `"meta"` to this field (e.g. [`"content"`, `"meta"`]).
            In this case the id will be generated by using the content and the defined metadata.
        """
        if known_ligatures is None: known_ligatures = KNOWN_LIGATURES
        if not isinstance(files, list): files = [files]
        prepared_files = [await self.async_prepare_file(file) for file in files]

        if isinstance(metadata, dict) or metadata is None: metadata = [metadata] * len(prepared_files)
        parsed_files: List[ParsedFile] = []

        for prepared_file, file_meta in self.get_iterator(prepared_files, metadata):
            parsed_files.extend(
                await self.async_convert(
                    file=prepared_file,
                    meta=file_meta,
                    remove_numeric_tables=remove_numeric_tables,
                    valid_languages=valid_languages,
                    encoding=encoding,
                    id_hash_keys=id_hash_keys,
                    extract_metadata=extract_metadata,
                    **kwargs,
                )
            )

        # Cleanup ligatures
        for parsed_file in parsed_files:
            for ligature, letters in known_ligatures.items():
                if parsed_file.content is not None: parsed_file.content = parsed_file.content.replace(ligature, letters)
        return parsed_files
    

    def run_batch(  # type: ignore
        self,
        files: Union['FileLike', 'PreparedFile', List['FileLike'], List['PreparedFile']],
        metadata: Optional[Union[Dict[str, str], List[Optional[Dict[str, str]]]]] = None,
        remove_numeric_tables: Optional[bool] = None,
        known_ligatures: Optional[Dict[str, str]] = None,
        valid_languages: Optional[List[str]] = None,
        encoding: Optional[str] = "UTF-8",
        id_hash_keys: Optional[List[str]] = None,
        extract_metadata: Optional[bool] = None,
        **kwargs,
    ) -> List[ParsedFile]:
        return self.run(
            files=files,
            metadata=metadata,
            remove_numeric_tables=remove_numeric_tables,
            known_ligatures=known_ligatures,
            valid_languages=valid_languages,
            encoding=encoding,
            id_hash_keys=id_hash_keys,
            extract_metadata=extract_metadata,
            **kwargs,
        )
    
    async def async_run_batch(  # type: ignore
        self,
        files: Union['FileLike', 'PreparedFile', List['FileLike'], List['PreparedFile']],
        metadata: Optional[Union[Dict[str, str], List[Optional[Dict[str, str]]]]] = None,
        remove_numeric_tables: Optional[bool] = None,
        known_ligatures: Optional[Dict[str, str]] = None,
        valid_languages: Optional[List[str]] = None,
        encoding: Optional[str] = "UTF-8",
        id_hash_keys: Optional[List[str]] = None,
        extract_metadata: Optional[bool] = None,
        **kwargs,
    ) -> List[ParsedFile]:
        return await self.async_run(
            files=files,
            metadata=metadata,
            remove_numeric_tables=remove_numeric_tables,
            known_ligatures=known_ligatures,
            valid_languages=valid_languages,
            encoding=encoding,
            id_hash_keys=id_hash_keys,
            extract_metadata=extract_metadata,
            **kwargs,
        )
    

    async def async_from_request(
        self,
        request: Optional['Request'] = None, 
        upload_files: Optional[Union[List['UploadFile'], 'UploadFile']] = None, 
        upload_form_keys: Optional[Union[List[str], str]] = None,
        remote_files: Optional[Union[List[str], str]] = None,
        remote_form_keys: Optional[List[str]] = None,

        metadata: Optional[Union[Dict[str, str], List[Optional[Dict[str, str]]]]] = None,
        remove_numeric_tables: Optional[bool] = None,
        known_ligatures: Optional[Dict[str, str]] = None,
        valid_languages: Optional[List[str]] = None,
        encoding: Optional[str] = "UTF-8",
        id_hash_keys: Optional[List[str]] = None,
        extract_metadata: Optional[bool] = None,
        **kwargs
    ) -> List[ParsedFile]:
        """
        Parse a request and extract text from the files.
        """
        prepared_files = await PreparedFile.async_from_request(
            request=request,
            upload_files=upload_files,
            upload_form_keys=upload_form_keys,
            remote_files=remote_files,
            remote_form_keys=remote_form_keys,
            **kwargs
        )
        return await self.async_run(
            files=prepared_files,
            metadata=metadata,
            remove_numeric_tables=remove_numeric_tables,
            known_ligatures=known_ligatures,
            valid_languages=valid_languages,
            encoding=encoding,
            id_hash_keys=id_hash_keys,
            extract_metadata=extract_metadata,
            **kwargs,
        )







"""
Imports for API Classes
"""

try:
    from fastapi import UploadFile as FastUploadFile
except ImportError:
    FastUploadFile = object

try:
    from starlette.datastructures import UploadFile as StarletteUploadFile
except ImportError:
    StarletteUploadFile = object

try:
    from starlite.datastructures import UploadFile as StarliteUploadFile
except ImportError:
    StarliteUploadFile = object


__all__ = [
    'FastUploadFile',
    'StarletteUploadFile',
    'StarliteUploadFile'
]


"""
Subclass for Wasabi S3
"""

import re
import contextlib
with contextlib.suppress(ImportError):
    import s3fs

bucket_format_list = [
    re.compile(
        r"^(?P<bucket>arn:(aws).*:wsbi:[a-z\-0-9]*:[0-9]{12}:accesspoint[:/][^/]+)/?"
        r"(?P<key>.*)$"
    ),
    re.compile(
        r"^(?P<bucket>arn:(aws).*:ws:[a-z\-0-9]*:[0-9]{12}:accesspoint[:/][^/]+)/?"
        r"(?P<key>.*)$"
    ),
    re.compile(
        r"^(?P<bucket>arn:(aws).*:s3:[a-z\-0-9]*:[0-9]{12}:accesspoint[:/][^/]+)/?"
        r"(?P<key>.*)$"
    ),
    re.compile(
        r"^(?P<bucket>arn:(aws).*:s3-outposts:[a-z\-0-9]+:[0-9]{12}:outpost[/:]"
        r"[a-zA-Z0-9\-]{1,63}[/:](bucket|accesspoint)[/:][a-zA-Z0-9\-]{1,63})[/:]?(?P<key>.*)$"
    ),
    re.compile(
        r"^(?P<bucket>arn:(aws).*:s3-outposts:[a-z\-0-9]+:[0-9]{12}:outpost[/:]"
        r"[a-zA-Z0-9\-]{1,63}[/:]bucket[/:]"
        r"[a-zA-Z0-9\-]{1,63})[/:]?(?P<key>.*)$"
    ),
    re.compile(
        r"^(?P<bucket>arn:(aws).*:s3-object-lambda:[a-z\-0-9]+:[0-9]{12}:"
        r"accesspoint[/:][a-zA-Z0-9\-]{1,63})[/:]?(?P<key>.*)$"
    ),
]

class WasabiFileSystem(s3fs.S3FileSystem):
    """
    Access Wasabi as if it were a file system.

    This exposes a filesystem-like API (ls, cp, open, etc.) on top of S3
    storage.

    Provide credentials either explicitly (``key=``, ``secret=``) or depend
    on boto's credential methods. See botocore documentation for more
    information. If no credentials are available, use ``anon=True``.

    Parameters
    ----------
    anon : bool (False)
        Whether to use anonymous connection (public buckets only). If False,
        uses the key/secret given, or boto's credential resolver (client_kwargs,
        environment, variables, config files, EC2 IAM server, in that order)
    endpoint_url : string (None)
        Use this endpoint_url, if specified. Needed for connecting to non-AWS
        S3 buckets. Takes precedence over `endpoint_url` in client_kwargs.
    key : string (None)
        If not anonymous, use this access key ID, if specified. Takes precedence
        over `aws_access_key_id` in client_kwargs.
    secret : string (None)
        If not anonymous, use this secret access key, if specified. Takes
        precedence over `aws_secret_access_key` in client_kwargs.
    token : string (None)
        If not anonymous, use this security token, if specified
    use_ssl : bool (True)
        Whether to use SSL in connections to S3; may be faster without, but
        insecure. If ``use_ssl`` is also set in ``client_kwargs``,
        the value set in ``client_kwargs`` will take priority.
    s3_additional_kwargs : dict of parameters that are used when calling s3 api
        methods. Typically used for things like "ServerSideEncryption".
    client_kwargs : dict of parameters for the botocore client
    requester_pays : bool (False)
        If RequesterPays buckets are supported.
    default_block_size: int (None)
        If given, the default block size value used for ``open()``, if no
        specific value is given at all time. The built-in default is 5MB.
    default_fill_cache : Bool (True)
        Whether to use cache filling with open by default. Refer to
        ``S3File.open``.
    default_cache_type : string ("readahead")
        If given, the default cache_type value used for ``open()``. Set to "none"
        if no caching is desired. See fsspec's documentation for other available
        cache_type values. Default cache_type is "readahead".
    version_aware : bool (False)
        Whether to support bucket versioning.  If enable this will require the
        user to have the necessary IAM permissions for dealing with versioned
        objects. Note that in the event that you only need to work with the
        latest version of objects in a versioned bucket, and do not need the
        VersionId for those objects, you should set ``version_aware`` to False
        for performance reasons. When set to True, filesystem instances will
        use the S3 ListObjectVersions API call to list directory contents,
        which requires listing all historical object versions.
    cache_regions : bool (False)
        Whether to cache bucket regions or not. Whenever a new bucket is used,
        it will first find out which region it belongs and then use the client
        for that region.
    asynchronous :  bool (False)
        Whether this instance is to be used from inside coroutines.
    config_kwargs : dict of parameters passed to ``botocore.client.Config``
    kwargs : other parameters for core session.
    session : aiobotocore AioSession object to be used for all connections.
         This session will be used inplace of creating a new session inside S3FileSystem.
         For example: aiobotocore.session.AioSession(profile='test_user')

    The following parameters are passed on to fsspec:

    skip_instance_cache: to control reuse of instances
    use_listings_cache, listings_expiry_time, max_paths: to control reuse of directory listings

    Examples
    --------
    >>> s3 = WasabiFileSystem(anon=False)  # doctest: +SKIP
    >>> s3.ls('my-bucket/')  # doctest: +SKIP
    ['my-file.txt']

    >>> with s3.open('my-bucket/my-file.txt', mode='rb') as f:  # doctest: +SKIP
    ...     print(f.read())  # doctest: +SKIP
    b'Hello, world!'
    """

    def _find_bucket_key(self, s3_path: str):
        """
        This is a helper function that given an s3 path such that the path is of
        the form: bucket/key
        It will return the bucket and the key represented by the s3 path
        """

        
        for bucket_format in bucket_format_list:
            match = bucket_format.match(s3_path)
            if match:
                return match.group("bucket"), match.group("key")
        s3_components = s3_path.split("/", 1)
        bucket = s3_components[0]
        s3_key = s3_components[1] if len(s3_components) > 1 else ""
        return bucket, s3_key


"""
Core Imports
""""""
Handles Importing the correct pathlib for this library in
python3.10 cases
"""

import os
import io
import sys
import inspect
import ntpath
import posixpath

from anyio import open_file, AsyncFile
from os import stat_result, PathLike
from contextlib import asynccontextmanager

from typing import TYPE_CHECKING, Union
from typing import Optional, List, AsyncIterable, Iterable, IO, AsyncContextManager, cast, Callable

from fileio.types import Final, Literal, FileMode

from fileio.lib.aiopath.wrap import coro_as_method_coro, func_as_method_coro, to_thread, method_as_method_coro, func_to_async_func
from fileio.lib.aiopath.handle import IterableAIOFile, get_handle


# if 3.10
if sys.version_info.minor >= 10:
    from fileio.utils.ops import register_pydantic_type
    import fileio.lib.pathz as pathlib
    from fileio.lib.pathz import PosixPath, WindowsPath, Path, PurePath, _ignore_error
    from fileio.lib.pathz import _NormalAccessor as NormalAccessor
    from fileio.lib.pathz import _make_selector as _sync_make_selector
    from fileio.lib.pathz import _PosixFlavour, _WindowsFlavour
    from fileio.lib.pathz import _getfinalpathname
    if _getfinalpathname is not None:
        _async_getfinalpathname = func_to_async_func(_getfinalpathname)
    else:
        def _getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")

        async def _async_getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")

    register_pydantic_type(Path, str)
    register_pydantic_type(PosixPath, str)
    register_pydantic_type(WindowsPath, str)

else:
    import pathlib
    from pathlib import PosixPath, WindowsPath, Path, PurePath, _ignore_error
    from pathlib import _NormalAccessor as NormalAccessor
    from pathlib import _make_selector as _sync_make_selector
    from pathlib import _PosixFlavour, _WindowsFlavour
    try:
        from pathlib import _getfinalpathname
        _async_getfinalpathname = func_to_async_func(_getfinalpathname)

    except ImportError:
        def _getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")
    
        async def _async_getfinalpathname(*args, **kwargs):
            raise ImportError("_getfinalpathname() requires a Windows/NT platform")



DEFAULT_ENCODING: Final[str] = 'utf-8'
ON_ERRORS: Final[str] = 'ignore'
NEWLINE: Final[str] = '\n'

BEGINNING: Final[int] = 0
CHUNK_SIZE: Final[int] = 4 * 1_024

SEP: Final[str] = '\n'
ENCODING: Final[str] = 'utf-8'
ERRORS: Final[str] = 'replace'

FileData = Union[bytes, str]
Paths = Union[Path, PathLike, str]
Handle = AsyncFile


def iscoroutinefunction(obj):
    if inspect.iscoroutinefunction(obj): return True
    return bool(hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__))


__all__ = (
    'io',
    'os',
    'inspect',
    'ntpath',
    'posixpath',
    'open_file',
    'AsyncFile',
    'stat_result',
    'PathLike',
    'asynccontextmanager',
    'pathlib', 
    'PosixPath',
    'WindowsPath',
    'PurePath',
    'Path',
    '_ignore_error',
    '_PosixFlavour',
    '_WindowsFlavour',
    'NormalAccessor',
    '_sync_make_selector',
    '_getfinalpathname',
    '_async_getfinalpathname',
    'coro_as_method_coro',
    'func_as_method_coro',
    'to_thread',
    'method_as_method_coro',
    'func_to_async_func',
    'IterableAIOFile',
    'Final',
    'Literal',
    'get_handle',
    'FileMode',
    'DEFAULT_ENCODING',
    'ON_ERRORS',
    'NEWLINE',
    'TYPE_CHECKING',
    'BEGINNING',
    'CHUNK_SIZE',
    'SEP',
    'ENCODING',
    'ERRORS',
    'FileData', 'Paths', 'Handle',
    'iscoroutinefunction',
    'Union', 'Optional', 'List', 'AsyncIterable', 'Iterable', 'IO', 'AsyncContextManager', 'cast', 'Callable'
)



from pathlib import Path
from typing import List, Dict, Union, Any, Optional

try:
    import ruamel.yaml
    import ruamel.yaml.scalarstring as yaml_scalarstring
    _yaml_available = True
except ImportError:
    _yaml_available = False


if _yaml_available:
    yaml = ruamel.yaml.YAML(typ=['rt', 'string'])
    yaml.preserve_quotes = True
    yaml.indent(
        mapping = 2, 
        sequence = 2, 
        offset = 0
    )


    class Yaml:

        @classmethod
        def dump(cls, obj: Dict[Any, Any], stream: Any = None, *args, **kwargs) -> Any:
            return yaml.dump(obj, stream = stream, *args, **kwargs)

        @classmethod
        def dumps(cls, obj: Union[List[Any], Dict[Any, Any]], stream: Any = None, *args, **kwargs) -> Any:
            return yaml.dump_to_string(obj, *args, **kwargs)
        
        @classmethod
        def _convert_str_types(cls, data: Union[List[Any], Dict[Any, Any], Any]) -> Union[List[Any], Dict[Any, Any]]:
            if isinstance(data, list):
                return [cls._convert_str_types(d) for d in data]
            if isinstance(data, dict):
                return {k: cls._convert_str_types(v) for k, v in data.items()}
            if isinstance(data, (yaml_scalarstring.ScalarString, yaml_scalarstring.SingleQuotedScalarString, yaml_scalarstring.DoubleQuotedScalarString)):
                return str(data)

        @classmethod
        def load(cls, data: Union[str, bytes, Any], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
            return yaml.load(data, *args, **kwargs)
        
        @classmethod
        def loads_multi(cls, data: Union[str, bytes, Any], *args, _seperator: str = '\n---', convert_types: Optional[bool] = False, **kwargs) -> Union[Dict[Any, Any], List[Dict[Any, Any]]]:
            """
            Loads a multi-yaml string into a list of Dict[Yaml].
            """
            _data = data.split(_seperator)
            try: 
                result = [yaml.load(d, *args, **kwargs) for d in _data]
                if convert_types: result = [cls._convert_str_types(d) for d in result]
            except Exception as e: 
                result = yaml.load(data, *args, **kwargs)
                if convert_types: result = cls._convert_str_types(result)
            return result
            

        @classmethod
        def loads(cls, data: Union[str, bytes, Path, Any], *args, multi: bool = False, convert_types: Optional[bool] = False, **kwargs) -> Union[Dict[Any, Any], List[str], List[Dict[Any, Any]]]:
            if not isinstance(data, str) and hasattr(data, 'as_posix'):
                # If it's a file-like object, read it as a string
                data = data.read_text()
            if '\n---' in data: multi = True 
            if multi: return cls.loads_multi(data, *args, convert_types = convert_types, **kwargs)
            data = yaml.load(data, *args, **kwargs)
            if convert_types: data = cls._convert_str_types(data)
            return data



else:

    class Yaml:
        @classmethod
        def dump(cls, obj: Dict[Any, Any], stream: Any = None, *args, **kwargs) -> Any:
            raise ImportError("ruamel.yaml is not installed.")

        @classmethod
        def dumps(cls, obj: Dict[Any, Any], stream: Any = None, *args, **kwargs) -> Any:
            raise ImportError("ruamel.yaml is not installed.")

        @classmethod
        def load(cls, data: Union[str, bytes, Any], *args, **kwargs) -> Union[Dict[Any, Any], List[str]]:
            raise ImportError("ruamel.yaml is not installed.")
        
        @classmethod
        def loads_multi(cls, data: Union[str, bytes, Any], *args, _seperator: str = '\n---', **kwargs) -> Union[Dict[Any, Any], List[Dict[Any, Any]]]:
            """
            Loads a multi-yaml string into a list of Dict[Yaml].
            """
            raise ImportError("ruamel.yaml is not installed.")
            

        @classmethod
        def loads(cls, data: Union[str, bytes, Path, Any], *args, multi: bool = False, **kwargs) -> Union[Dict[Any, Any], List[str], List[Dict[Any, Any]]]:
            raise ImportError("ruamel.yaml is not installed.")
            



"""
Additional File Helpers
"""
import os
import aiohttpx
import hashlib
import contextlib

from typing import Callable, Dict, Any, Tuple, Optional, TYPE_CHECKING
from .helpers import timed_cache
from .configs import settings

if TYPE_CHECKING:
    from fileio.lib.types import FileLike
    with contextlib.suppress(ImportError):
        from starlette.requests import Request
        from starlette.datastructures import UploadFile


def get_url_file_name(url: str):
    url = url.split("#")[0]
    url = url.split("?")[0]
    return os.path.basename(url)

@timed_cache(secs = settings.checksum_cache_ttl)
def checksum_file(path: 'FileLike', chunk_size: Optional[int] = None) -> str:
    """
    Takes the uploaded file from the request and performs
    a SHA256 checksum to determine the uniqueness of the file.

    :param path: The path to the file
    :param chunk_size: The size of the chunks to read from the file (default is 64kb)
    """
    from fileio.lib.types import File

    if not chunk_size: chunk_size = settings.read_chunk_size
    path = File(path)
    sha256_hash = hashlib.sha256()
    with path.open('rb') as f:
        for byte_block in iter(lambda: f.read(chunk_size), b""):
            sha256_hash.update(byte_block)
    checksum = sha256_hash.hexdigest()
    del sha256_hash
    return checksum


@timed_cache(secs = settings.checksum_cache_ttl)
async def async_checksum_file(path: 'FileLike', chunk_size: Optional[int] = None) -> str:
    """
    [Async]
    Takes the uploaded file from the request and performs
    a SHA256 checksum to determine the uniqueness of the file.

    :param path: The path to the file
    :param chunk_size: The size of the chunks to read from the file (default is 64kb)
    """
    from fileio.lib.types import File

    if not chunk_size: chunk_size = settings.read_chunk_size
    path = File(path)
    sha256_hash = hashlib.sha256()
    async with path.async_open('rb') as f:
        if not path.is_cloud:
            for byte_block in await iter(lambda: f.read(chunk_size), b""):
                sha256_hash.update(byte_block)
        else:
            byte_block = await f.read(chunk_size)
            while byte_block:
                sha256_hash.update(byte_block)
                byte_block = await f.read(chunk_size)
            
    checksum = sha256_hash.hexdigest()
    del sha256_hash
    return checksum

def fetch_file_from_url(
    url: str,
    path: Optional['FileLike'] = None,
    directory: Optional[str] = None,
    filename: Optional[str] = None,
    chunk_size: Optional[int] = None,
    request_kwargs: Optional[Dict[str, str]] = None,
    overwrite: Optional[bool] = False,
    **kwargs,
) -> 'FileLike':
    """
    Fetches a file from a URL and saves it to the specified path.
    """
    from fileio.lib.types import File

    assert '://' in url, f'Invalid URL: {url}'
    if not chunk_size: chunk_size = settings.url_chunk_size
    if not path and not filename and not directory:
        path = File.get_tempfile()
    elif filename or directory:
        path_dir = File(directory) if directory else File.get_tempdir()
        path_dir.mkdir()
        path = path_dir.joinpath(filename or get_url_file_name(url))
    else:
        path = File(path)
    if not overwrite and path.exists(): return path
    with path.open('wb') as f:
        client = aiohttpx.Client()
        request_kwargs = request_kwargs or {}
        with client.stream('GET', url, **request_kwargs) as resp:
            for chunk in resp.aiter_bytes(chunk_size = chunk_size):
                f.write(chunk)
    return path

async def async_fetch_file_from_url(
    url: str,
    path: Optional['FileLike'] = None,
    directory: Optional[str] = None,
    filename: Optional[str] = None,
    chunk_size: Optional[int] = None,
    request_kwargs: Optional[Dict[str, str]] = None,
    overwrite: Optional[bool] = False,
    **kwargs,
) -> 'FileLike':
    """
    Fetches a file from a URL and saves it to the specified path.
    """
    from fileio.lib.types import File

    assert '://' in url, f'Invalid URL: {url}'

    if not chunk_size: chunk_size = settings.url_chunk_size
    if not path and not filename and not directory:
        path = File.get_tempfile()
    elif filename or directory:
        path_dir = File(directory) if directory else File.get_tempdir()
        await path_dir.async_mkdir()
        path = path_dir.joinpath(filename or get_url_file_name(url))
    else:
        path = File(path)
    if not overwrite and await path.async_exists(): return path
    async with path.async_open('wb') as f:
        client = aiohttpx.Client()
        request_kwargs = request_kwargs or {}
        async with client.async_stream('GET', url, **request_kwargs) as resp:
            async for chunk in resp.aiter_bytes(chunk_size = chunk_size):
                await f.write(chunk)
    return path


try:
    from pydantic.json import ENCODERS_BY_TYPE
except ImportError:
    ENCODERS_BY_TYPE = None 


def register_pydantic_type(obj, t):
    """
    Registers the type with Pydantic's JSON encoder.
    """
    global ENCODERS_BY_TYPE
    if ENCODERS_BY_TYPE is None: return
    # print(f"Registering {obj} = {t} with Pydantic's JSON encoder.")
    ENCODERS_BY_TYPE[obj] = t



import datetime
import contextlib
from typing import Any, Dict, List, Optional, Union, TYPE_CHECKING
from pydantic import BaseModel, Field, validator
from .classprops import lazyproperty


if TYPE_CHECKING:
    from fileio.lib.types import FileLike
    with contextlib.suppress(ImportError):
        from starlette.requests import Request
        from starlette.datastructures import UploadFile


class FileInfo(BaseModel):
    size: Optional[int] = None
    etag: Optional[str] = None
    last_modified: Optional[datetime.datetime] = None
    checksum: Optional[str] = None
    path: Optional['FileLike'] = None

    class Config:
        allow_arbitrary_types = True

    def dict(self, *args, **kwargs):
        d = super().dict(*args, **kwargs)
        if d['path']: d['path'] = self.path.as_posix()
        return d

    def validate_info(self):
        """
        Ensures that the file info is valid.
        """
        if not self.path or not self.path.exists():
            raise ValueError('Path does not exist.')
        if self.checksum is None:
            from fileio.utils.ops import checksum_file
            self.checksum = checksum_file(self.path)
        if not all(self.size, self.etag, self.last_modified):
            info: Dict[str, Any] = self.path.info()
            self.size = info.get('size', info.get('Size', 0))
            self.etag = info.get('ETag') or 'none'
            self.last_modified = info.get('LastModified')
        
    async def async_validate_info(self):
        """
        Ensures that the file info is valid.
        """
        if not self.path or not await self.path.async_exists():
            raise ValueError('Path does not exist.')
        if self.checksum is None:
            from fileio.utils.ops import async_checksum_file
            self.checksum = await async_checksum_file(self.path)
        if not all(self.size, self.etag, self.last_modified):
            info: Dict[str, Any] = await self.path.async_info()
            self.size = info.get('size', info.get('Size', 0))
            self.etag = info.get('ETag') or 'none'
            self.last_modified = info.get('LastModified')

    @classmethod
    def get_info(cls, path: 'FileLike') -> 'FileInfo':
        """
        Fetches file info from the given path.
        """
        from fileio import File
        from fileio.utils.ops import checksum_file

        path = File(path)
        file_info: Dict[str, Any] = path.info()
        return cls(
            size = file_info.get('size', file_info.get('Size', 0)),
            etag = file_info.get('ETag') or 'none',
            last_modified = file_info.get('LastModified'),
            checksum = checksum_file(path = path),
            path = path
        )
    
    @classmethod
    async def async_get_info(cls, path: 'FileLike') -> 'FileInfo':
        """
        Fetches file info from the given path.
        """
        from fileio import File
        from fileio.utils.ops import async_checksum_file
        
        path = File(path)
        file_info: Dict[str, Any] = await path.async_info()
        return cls(
            size = file_info.get('size', file_info.get('Size', 0)),
            etag = file_info.get('ETag') or 'none',
            last_modified = file_info.get('LastModified'),
            checksum = await async_checksum_file(path = path),
            path = path
        )

    def __eq__(self, other):
        if isinstance(other, FileInfo):
            return self.checksum == other.checksum
        return False

class FileInfo(BaseModel):
    size: Optional[int] = None
    etag: Optional[str] = None
    last_modified: Optional[datetime.datetime] = None
    checksum: Optional[str] = None
    path: Optional['FileLike'] = None

    class Config:
        allow_arbitrary_types = True

    def dict(self, *args, **kwargs):
        d = super().dict(*args, **kwargs)
        if d['path']: d['path'] = self.path.as_posix()
        return d

    def validate_info(self):
        """
        Ensures that the file info is valid.
        """
        if not self.path or not self.path.exists():
            raise ValueError('Path does not exist.')
        if self.checksum is None:
            from fileio.utils.ops import checksum_file
            self.checksum = checksum_file(self.path)
        if not all(self.size, self.etag, self.last_modified):
            info: Dict[str, Any] = self.path.info()
            self.size = info.get('size', info.get('Size', 0))
            self.etag = info.get('ETag') or 'none'
            self.last_modified = info.get('LastModified')
        
    async def async_validate_info(self):
        """
        Ensures that the file info is valid.
        """
        if not self.path or not await self.path.async_exists():
            raise ValueError('Path does not exist.')
        if self.checksum is None:
            from fileio.utils.ops import async_checksum_file
            self.checksum = await async_checksum_file(self.path)
        if not all(self.size, self.etag, self.last_modified):
            info: Dict[str, Any] = await self.path.async_info()
            self.size = info.get('size', info.get('Size', 0))
            self.etag = info.get('ETag') or 'none'
            self.last_modified = info.get('LastModified')

    @classmethod
    def get_info(cls, path: 'FileLike') -> 'FileInfo':
        """
        Fetches file info from the given path.
        """
        from fileio import File
        from fileio.utils.ops import checksum_file

        path = File(path)
        file_info: Dict[str, Any] = path.info()
        return cls(
            size = file_info.get('size', file_info.get('Size', 0)),
            etag = file_info.get('ETag') or 'none',
            last_modified = file_info.get('LastModified'),
            checksum = checksum_file(path = path),
            path = path
        )
    
    @classmethod
    async def async_get_info(cls, path: 'FileLike') -> 'FileInfo':
        """
        Fetches file info from the given path.
        """
        from fileio import File
        from fileio.utils.ops import async_checksum_file
        
        path = File(path)
        file_info: Dict[str, Any] = await path.async_info()
        return cls(
            size = file_info.get('size', file_info.get('Size', 0)),
            etag = file_info.get('ETag') or 'none',
            last_modified = file_info.get('LastModified'),
            checksum = await async_checksum_file(path = path),
            path = path
        )

    def __eq__(self, other):
        if isinstance(other, FileInfo):
            return self.checksum == other.checksum
        return False



class PreparedFile(BaseModel):

    local_path: Optional[Union[Any, 'FileLike']] = None
    remote_path: Optional[Union[Any, 'FileLike']] = None
    is_tmp: Optional[bool] = None
    checksum: Optional[str] = None

    @lazyproperty
    def is_temp(self):
        return self.is_tmp or (self.local_path is not None and 'tmp' in self.local_path.as_posix())
    
    @lazyproperty
    def source_path(self) -> 'FileLike':
        return self.local_path or self.remote_path
    
    @lazyproperty
    def file_info(self) -> FileInfo:
        return FileInfo.get_info(self.source_path)
    
    def as_posix(self, **kwargs):
        return self.source_path.as_posix(**kwargs)
    
    def to_posix(
        self, 
        local_required: Optional[bool] = False, 
        remote_required: Optional[bool] = False,
        **kwargs,
    ) -> str:
        """
        Returns the path as a posix path.
        """
        if remote_required:
            if not self.remote_path:
                raise ValueError(f'Remote path is required and source path: {self.source_path} is not remote.')
            return self.remote_path.as_posix(**kwargs)
        if local_required:
            if not self.local_path and not self.remote_path:
                raise ValueError(f'Local path is required and source path: {self.source_path} is not local.')
            if not self.local_path:
                from fileio import File
                self.local_path = self.remote_path.copy_file(dest = File.get_tempfile())
            return self.local_path.as_posix(**kwargs)
        return self.source_path.as_posix(**kwargs)
        # raise ValueError(f'Local path is required and source path: {self.source_path} is not local.')
    
    async def async_to_posix(
        self, 
        local_required: Optional[bool] = False, 
        remote_required: Optional[bool] = False,
        **kwargs,
    ) -> str:
        """
        Returns the path as a posix path.
        """
        if remote_required:
            if not self.remote_path:
                raise ValueError(f'Remote path is required and source path: {self.source_path} is not remote.')
            return self.remote_path.as_posix(**kwargs)
        if local_required:
            if not self.local_path and not self.remote_path:
                raise ValueError(f'Local path is required and source path: {self.source_path} is not local.')
            if not self.local_path:
                from fileio import File
                self.local_path = await self.remote_path.async_copy_file(dest = File.get_tempfile())
            return self.local_path.as_posix(**kwargs)
        return self.source_path.as_posix(**kwargs)
    
    @classmethod
    def from_path(cls, path: 'FileLike', is_tmp: Optional[bool] = None, **kwargs):
        from fileio import File
        from fileio.utils.ops import checksum_file
        p = File(path)
        cksum = checksum_file(p)
        kws = {
            'local_path': None if p.is_cloud else p,
            'remote_path': p if p.is_cloud else None,
            'is_tmp': is_tmp if is_tmp is not None else (not p.is_cloud and p.is_temp),
            'checksum': cksum,
            **kwargs,
        }
        return cls.parse_obj(kws)
    
    @classmethod
    async def async_from_path(cls, path: 'FileLike', is_tmp: Optional[bool] = None, **kwargs):
        from fileio import File
        from fileio.utils.ops import async_checksum_file
        p = File(path)
        cksum = await async_checksum_file(p)
        kws = {
            'local_path': None if p.is_cloud else p,
            'remote_path': p if p.is_cloud else None,
            'is_tmp': is_tmp if is_tmp is not None else (not p.is_cloud and p.is_temp),
            'checksum': cksum,
            **kwargs,
        }
        return cls.parse_obj(kws)
    
    @classmethod
    def from_url(cls, url, **kwargs):
        from fileio.utils.ops import checksum_file, fetch_file_from_url
        p = fetch_file_from_url(url, **kwargs)
        cksum = checksum_file(p)
        kws = {
            'local_path': None if p.is_cloud else p,
            'remote_path': p if p.is_cloud else None,
            'is_tmp': p.is_temp,
            'checksum': cksum,
            **kwargs,
        }
        return cls.parse_obj(kws)
    
    @classmethod
    async def async_from_url(cls, url, **kwargs):
        from fileio.utils.ops import async_checksum_file, async_fetch_file_from_url
        p = await async_fetch_file_from_url(url, **kwargs)
        cksum = await async_checksum_file(p)
        kws = {
            'local_path': None if p.is_cloud else p,
            'remote_path': p if p.is_cloud else None,
            'is_tmp': p.is_temp,
            'checksum': cksum,
            **kwargs,
        }
        return cls.parse_obj(kws)

    def dict(self, *args, **kwargs):
        d = super().dict(*args, **kwargs)
        if d['local_path']: d['local_path'] = self.local_path.as_posix()
        if d['remote_path']: d['remote_path'] = self.remote_path.as_posix()
        d['source_path'] = self.source_path.as_posix()
        d['is_tmp'] = d.pop('is_temp', False)
        return d

    @classmethod
    async def async_from_remote(
        cls,
        request: Optional['Request'] = None, 
        remote_files: Optional[Union[List[str], str]] = None,
        remote_form_keys: Optional[List[str]] = None,
        **kwargs
    ) -> Union['PreparedFile', List['PreparedFile']]:
        """
        Create a PreparedFile from a remote file(s)
        """
        if not remote_files and request:
            form_data = await request.form()
            remote_files: List[str] = []
            if not remote_form_keys:
                remote_form_keys = ['url', 'urls', 'remote', 'remote_file', 'remote_files']
                for k in remote_form_keys:
                    if form_data.get(k):
                        if isinstance(form_data[k], list):
                            remote_files.extend(form_data[k])
                        else:
                            remote_files.append(form_data[k])

        if not remote_files: raise ValueError("No remote file(s) found")
        if not isinstance(remote_files, list): remote_files = [remote_files]
        
        
        from fileio import File
        from fileio.utils.ops import async_checksum_file, async_fetch_file_from_url
        files = []
        for f in remote_files:
            if 'http' in f: remote_file = await async_fetch_file_from_url(url = f, **kwargs)
            else: remote_file = File(f)
            cksum = await async_checksum_file(remote_file)
            kws = {
                'local_path': None if remote_file.is_cloud else remote_file,
                'remote_path': remote_file if remote_file.is_cloud else None,
                'is_tmp': remote_file.is_temp,
                'checksum': cksum,
                **kwargs,
            }
            files.append(cls.parse_obj(kws))
        return files if len(files) > 1 else files[0]
    
    @classmethod
    async def async_from_upload(
        cls, 
        request: Optional['Request'] = None, 
        upload_files: Optional[Union[List['UploadFile'], 'UploadFile']] = None, 
        upload_form_keys: Optional[Union[List[str], str]] = None,
        **kwargs
    ) -> Union['PreparedFile', List['PreparedFile']]:
        """
        Create a PreparedFile from a request or upload_file
        """
        if not upload_files and request:
            form_data = await request.form()
            upload_files: List['UploadFile'] = []
            if not upload_form_keys:
                upload_form_keys = ['file', 'files', 'upload', 'upload_file', 'upload_files']
                for k in upload_form_keys:
                    if form_data.get(k):
                        if isinstance(form_data[k], list):
                            upload_files.extend(form_data[k])
                        else:
                            upload_files.append(form_data[k])

        if not upload_files: raise ValueError("No upload file(s) found")
        if not isinstance(upload_files, list): upload_files = [upload_files]

        from fileio import File
        from fileio.utils.ops import async_checksum_file
        from starlette.datastructures import UploadFile

        files = []
        for f in upload_files:
            _is_tmp = False
            if isinstance(f, UploadFile):
                file: FileLike = File.get_tempfile()
                await file.async_write_bytes(await f.read())
                _is_tmp = True
            else: file = File(f)
            cksum = await async_checksum_file(file)
            kws = {
                'local_path': None if file.is_cloud else file,
                'remote_path': file if file.is_cloud else None,
                'is_tmp': _is_tmp or file.is_temp,
                'checksum': cksum,
                **kwargs,
            }
            files.append(cls.parse_obj(kws))
        return files if len(files) > 1 else files[0]


    @classmethod
    async def async_from_request(
        cls, 
        request: Optional['Request'] = None, 
        upload_files: Optional[Union[List['UploadFile'], 'UploadFile']] = None, 
        upload_form_keys: Optional[Union[List[str], str]] = None,
        remote_files: Optional[Union[List[str], str]] = None,
        remote_form_keys: Optional[List[str]] = None,
        **kwargs
    ) -> Union['PreparedFile', List['PreparedFile']]:
        """
        Create a PreparedFile from a request or upload_file
        """
        if remote_files or remote_form_keys:
            return await cls.async_from_remote(request = request, remote_files = remote_files, remote_form_keys = remote_form_keys, **kwargs)
        return await cls.async_from_upload(request = request, upload_files = upload_files, upload_form_keys = upload_form_keys, **kwargs)

    def read_bytes(self, *args, **kwargs):
        return self.source_path.read_bytes(*args, **kwargs)
    
    def read_text(self, *args, **kwargs):
        return self.source_path.read_text(*args, **kwargs)
    
    async def async_read_bytes(self, *args, **kwargs):
        return await self.source_path.async_read_bytes(*args, **kwargs)
    
    async def async_read_text(self, *args, **kwargs):
        return await self.source_path.async_read_text(*args, **kwargs)

    class Config:
        allow_arbitrary_types = True
        extra = 'allow'


class ParsedFile(BaseModel):
    content: Union[str, Dict[str, Any], Any]
    metadata: Optional[Dict[str, Any]] = {}
    file: Optional[Union[PreparedFile, Any]] = Field(default_factory = PreparedFile)

    @lazyproperty
    def file_info(self) -> FileInfo:
        return self.file.file_info

    def dict(self, *args, **kwargs):
        d = super().dict(*args, **kwargs)
        if d['file']: d['file'] = self.file.dict()
        return d

    class Config:
        allow_arbitrary_types = True
        extra = 'allow'



class UniqueFile(BaseModel):

    filename: Optional[str] = Field(None, description = "The filename of the file")
    checksum: Optional[str] = Field(None, description = "The checksum of the file")
    checksum_method: Optional[str] = Field(None, description = "The checksum method of the file")
    mime_type: Optional[str] = Field(None, description = "The mime type of the file")
    extensions: Optional[Union[List[str], str]] = Field(None, description = "The extensions of the file")
    size: Optional[int] = Field(None, description = "The size of the file")
    etag: Optional[str] = Field(None, description = "The etag of the file")
    last_modified: Optional[datetime.datetime] = Field(None, description = "The last modified date of the file")

    @validator('extensions', pre = True)
    def _validate_extensions(cls, v):
        return [v] if isinstance(v, str) else v

    class Config:
        allow_arbitrary_types = True
        extra = 'allow'
    
    @property
    def id(self):
        return self.checksum
    
    def __eq__(self, other: Union[str, 'UniqueFile']):
        """
        Compare two UniqueFile objects
        """
        if isinstance(other, str): return self.id == other
        return (self.id == other.id and self.checksum_method == other.checksum_method)
    
    def __lt__(self, other: Union[int, float, 'UniqueFile']):
        """
        Compare two UniqueFile objects based on file size
        """
        if isinstance(other, (int, float)): return self.size < other
        return self.size < other.size

    def __gt__(self, other: Union[int, float, 'UniqueFile']):
        """
        Compare two UniqueFile objects based on file size
        """
        if isinstance(other, (int, float)): return self.size > other
        return self.size > other.size
    
    def __le__(self, other: Union[int, float, 'UniqueFile']):
        """
        Compare two UniqueFile objects based on file size
        """
        if isinstance(other, (int, float)): return self.size <= other
        return self.size <= other.size
    
    def __ge__(self, other: Union[int, float, 'UniqueFile']):
        """
        Compare two UniqueFile objects based on file size
        """
        if isinstance(other, (int, float)): return self.size >= other
        return self.size >= other.size
    





from typing import Dict
from .base import BasePack
from .base64 import Base64
from .base64gzip import Base64GZip
from ._json import Json, OrJson
from .pickle import Pickle, Dill
from .msgpack import MsgPack
from .generator import Generate
from .text import Text
from .yaml import Yaml
from .csv import Csv
from .tsv import Tsv
from .jsonlines import JsonLines

Serializers: Dict[str, BasePack] = {
    'pickle': Pickle,
    'dill': Dill,
    'json': Json,
    'orjson': OrJson,
    'base64': Base64,
    'bgzip': Base64GZip,
    'msgpack': MsgPack,
    'yaml': Yaml,
    'text': Text,
    'csv': Csv,
    'tsv': Tsv,
    'jsonlines': JsonLines,
    
}




from __future__ import annotations

"""
FileIO CLI Handler
"""

import typer
import base64
from pydantic.types import ByteSize
from fileio.lib.types import File
from lazyops.utils.helpers import timer
from typing import Optional, List, Union, Dict, Any, Callable, TYPE_CHECKING

async def fetch_file(
    source: str,
    destination: str,
    filename: Optional[str] = None,
    new_suffix: Optional[str] = None,
    overwrite: Optional[bool] = False,
) -> File:
    """
    Uploads a file using FileIO backend

    Source: Path to file to upload
    Destination: Path to upload file to. Can be a directory or a file
    Filename: Name of file to upload as. If not provided, will use the name of the source file
    NewSuffix: If provided, will replace the suffix of the source file with this suffix
    Overwrite: Whether to overwrite the destination file if it exists
    """

    source_path = File(source)
    destination_path = File(destination)

    assert await source_path.async_exists(), f'File {source_path} does not exist'
    assert await source_path.async_is_file(), f'File {source_path} is not a file'
    if await destination_path.async_exists() and not overwrite:
        raise FileExistsError(f'File {destination_path} already exists')
    
    t = timer()
    if destination_path.is_dir():
        await destination_path.async_mkdir(parents = True, exist_ok = True)
        filename = filename or source_path.name
        destination_path = destination_path.joinpath(filename)
        if new_suffix:
            destination_path = destination_path.with_suffix(new_suffix)
    
    await source_path.async_copy(destination_path, overwrite = overwrite)
    typer.echo(f'Copied {source_path} to {destination_path} in {timer(t):.2f} secs')
    return destination_path


async def source_info_file(
    source: str,
) -> Dict[str, Any]:
    """
    Returns information about a file or directory using FileIO backend
    """
    source_path = File(source)
    if not await source_path.async_exists():
        raise FileNotFoundError(f'File {source_path} does not exist')
    
    t = timer()
    stat = await source_path.async_info()
    # typer.echo(f'Got stats for {source_path} in {timer(t):.2f} secs')
    for key, value in stat.items():
        typer.echo(f'{key}: {value}')
    
    typer.echo(f'Retrieved stats for {source_path} in {timer(t):.2f} secs')
    return stat


async def source_info(
    source: str,
    recursive: Optional[bool] = False,
    dirs_only: Optional[bool] = False,
) -> Union[List[File], List[Dict[str, Any]]]:
    """
    Returns information about a file or directory using FileIO backend
    """

    source_path = File(source)
    if not await source_path.async_exists():
        raise FileNotFoundError(f'File {source_path} does not exist')
    if not await source_path.async_is_dir():
        return await source_info_file(source)

        # typer.echo(f'File {source_path} is not a directory. Listing parent directory instead')
        # source_path = source_path.parent

    t = timer()
    stats = await source_path.async_ls(
        recursive = recursive, 
        detail = True,
        prettify = True,
        as_path = False,
        files_only = not dirs_only,
    )
    # typer.echo(f'Retrieved stats for {source_path} in {timer(t):.2f} secs')

    total_size = sum(stat['Size'] for stat in stats)
    total_size_pretty = ByteSize.validate(total_size).human_readable()
    typer.echo(f'Source: {source_path}')
    for stat in stats:
        # typer.echo(f'Key: {stat["Key"]}')
        typer.echo(f'{stat["Key"]}')
        if stat.get('SizePretty'):
            typer.echo(f'- Size: {stat["SizePretty"]} ({stat["Size"]} bytes)')
        else:
            typer.echo(f'- Size: {ByteSize.validate(stat["Size"]).human_readable()} ({stat["Size"]} bytes)')
        if stat.get('LastModified'):
            typer.echo(f'- LastModified: {stat["LastModified"]}')
        if stat.get('ETag'):
            typer.echo(f'- ETag: {stat["ETag"]}')
        typer.echo('---' * 10)
        # typer.echo(f'Stats: {stat}')

    typer.echo(f'Total files: {len(stats)}, Total size: {total_size_pretty} ({total_size} bytes) (retrieved in {timer(t):.2f} secs)')
    return stats


async def list_dir(
    source: str,
    pattern: Optional[str] = '*',
) -> List[File]:
    """
    Lists a directory using FileIO backend
    """

    source_path = File(source)
    if not await source_path.async_exists():
        raise FileNotFoundError(f'File {source_path} does not exist')
    if not await source_path.async_is_dir():
        typer.echo(f'File {source_path} is not a directory. Listing parent directory instead')
        source_path = source_path.parent
    
    t = timer()
    files: List[File] = list(await source_path.async_glob(pattern))
    
    for file in files:
        kind = 'File' if await file.async_is_file() else 'Directory'
        typer.echo(f'- {file} ({kind})')
    
    typer.echo(f'Total Files: {len(files)} (retrieved in {timer(t):.2f} secs)')
    return files
    

async def source_exists(
    source: str,
) -> bool:
    """
    Checks if a file or directory exists using FileIO backend
    """

    source_path = File(source)
    exists = await source_path.async_exists()
    typer.echo(f'File {source_path} {"exists" if exists else "does not exist"}')
    return exists



async def rm_files_or_dir(
    source: str,
    recursive: Optional[bool] = False,
    force: Optional[bool] = False,
) -> None:
    """
    Removes a file or directory using FileIO backend
    """

    source_path = File(source)
    if not await source_path.async_exists():
        raise FileNotFoundError(f'File {source_path} does not exist')
    
    t = timer()
    if await source_path.async_is_dir():
        source_files = await source_path.async_ls(
            recursive = recursive, 
            detail = False,
            prettify = False,
            as_path = True,
            files_only = True,
        )
        typer.echo(f'Removing directory {source_path} with {len(source_files)} files (recursive: {recursive}, force: {force})')
        n_removed = 0
        if not force:
            typer.echo('Force is False. Will prompt for confirmation for each file. Use --force to skip confirmation')
        for source_file in source_files:
            if not force and not typer.confirm(f'Remove file {source_file}?'):
                continue
            await source_file.async_rm_file()
            n_removed += 1
        typer.echo(f'Removed {n_removed}/{len(source_files)} files in {timer(t):.2f} secs')

    
    else:
        await source_path.async_rm_file()
        typer.echo(f'Removed file {source_path} in {timer(t):.2f} secs')

# encode = lambda x: base64.b64encode(x).decode()

_supported_encodings: Dict[str, Dict[str, Union[str, Callable]]] = {
    'base64': {
        'encode': lambda x: base64.b64encode(x).decode('utf-8'),
        'mode': 'rb',
        'binary': True,
        'results': b'',
    }
}

_supported_decodings: Dict[str, Dict[str, Union[str, Callable]]] = {
    'base64': {
        'decode': lambda x: base64.b64decode(x),
        'mode': 'r',
        'binary': False,
        'results': '',
    }
}

async def cat_file(
    source: str,
    binary: Optional[bool] = False,
    encode: Optional[str] = None,
    decode: Optional[str] = None,
) -> Union[str, bytes]:
    """
    Extracts the contents of a file using FileIO backend
    """
    source_path = File(source)
    if not await source_path.async_exists():
        raise FileNotFoundError(f'File {source_path} does not exist')
    if not await source_path.async_is_file():
        raise FileNotFoundError(f'File {source_path} is not a file')
    
    if encode:
        assert encode in _supported_encodings, f'Encoding {encode} not supported: {_supported_encodings.keys()}'
        post_func = _supported_encodings[encode]['encode']
        mode = _supported_encodings[encode]['mode']
        results = _supported_encodings[encode]['results']
    elif decode:
        assert decode in _supported_decodings, f'Decoding {decode} not supported: {_supported_decodings.keys()}'
        post_func = _supported_decodings[decode]['decode']
        mode = _supported_decodings[decode]['mode']
        results = _supported_decodings[decode]['results']

    else:
        post_func = None
        mode = 'rb' if binary else 'r'
        results = b'' if binary else ''
    async with source_path.async_open(mode = mode) as f:
        async for line in f:
            results += line
    
    if post_func: results = post_func(results)
    typer.echo(results)
    return results


from __future__ import annotations

import os
import contextlib
from fileio.lib.posix.base import *
from fileio.lib.posix.filesys import R2_CloudFileSystem, R2_Accessor
from fileio.lib.posix.cloud import *
from fileio.utils import logger
from typing import Mapping, TYPE_CHECKING

if TYPE_CHECKING:
    from fileio.lib.types import FileLike

class FileR2PurePath(CloudFileSystemPurePath):
    _prefix: str = 'r2'
    _provider: str = 'CloudFlare'
    _win_pathz: ClassVar = 'PureFileR2WindowsPath'
    _posix_pathz: ClassVar = 'PureFileR2PosixPath'


class PureFileR2PosixPath(PureCloudFileSystemPosixPath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_posix_flavour
    _pathlike = posixpath
    __slots__ = ()


class PureFileR2WindowsPath(PureCloudFileSystemWindowsPath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _pathz_windows_flavour
    _pathlike = ntpath
    __slots__ = ()


class FileR2Path(CloudFileSystemPath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _pathz_windows_flavour if os.name == 'nt' else _pathz_posix_flavour
    _accessor: R2_Accessor = None
    _pathlike = posixpath
    _prefix = 'r2'
    _provider = 'CloudFlare'

    _win_pathz: ModuleType = 'FileR2WindowsPath'
    _posix_pathz: ModuleType = 'FileR2PosixPath'

    def _init(self, template: Optional['FileR2Path'] = None):
        self._accessor: R2_Accessor = FileSysManager.get_accessor(self._prefix)
        # self._accessor: AccessorLike = get_accessor(self._prefix)
        self._closed = False
        self._fileio = None

    def __new__(cls, *parts, **kwargs):
        if cls is FileR2Path or issubclass(cls, FileR2Path): 
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        self = cls._from_parts(parts, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self
    
    def async_open(self, mode: FileMode = 'r', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, block_size: int = 10 * 1024 * 1024, compression: str = None, **kwargs: Any) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        compression = infer doesn't work all that well.
        """
        #self._fileio = self._accessor.open(self._cloudpath, mode=mode, encoding=encoding, errors=errors, block_size=block_size, compression=compression, newline=newline, buffering=buffering, **kwargs)
        #print(type(self._fileio))
        #return get_cloud_file(self._fileio)
        # Test v2
        return get_cloudfs_file(
            self._accessor,
            self._cloudpath, 
            mode=mode, 
            encoding=encoding, 
            errors=errors, 
            block_size=block_size, 
            compression=compression, 
            newline=newline, 
            buffering=buffering, 
            **kwargs
        )
        # return get_cloud_file(self._accessor.open(self._cloudpath, mode=mode, encoding=encoding, errors=errors, block_size=block_size, compression=compression, newline=newline, buffering=buffering, **kwargs))


    # Implement some stuff that boto is faster in
    def upload_file(self, dest: 'FileLike', filename: Optional[str] = None, overwrite: bool = True, **kwargs):
        """
        Upload a file to R2

        Utilize boto3
        """
        if not overwrite and dest.exists(): raise FileExistsError(f"{dest} already exists and overwrite is False")
        filename = filename or self.name
        self._accessor.boto.upload_file(
            Bucket = self._bucket,
            Key = self.get_path_key(filename),
            Filename = dest.as_posix(),
        )
        return self.parent.joinpath(filename)
    
    def download_file(
        self, 
        output_file: Optional['FileLike'] = None,
        output_dir: Optional['FileLike'] = None,
        filename: Optional[str] = None,
        overwrite: bool = True,
        callbacks: Optional[List[Any]] = None,
        **kwargs
        ):
        """
        Downloads a file from R2 to a path
        """
        assert output_file or output_dir, "Must provide either output_file or output_dir"
        output_file = output_file or output_dir.joinpath(filename or self.name)
        assert overwrite or not output_file.exists(), f"{output_file} already exists and overwrite is False"
        # s3t = self._accessor.s3t()
        self._accessor.s3t.download(
            self._bucket,
            self.get_path_key(self.name),
            output_file.as_posix(),
            subscribers = callbacks
        )
        # self._accessor.s3t.shutdown()
        return output_file

    async def async_upload_file(
        self, 
        source: 'FileLike', 
        filename: Optional[str] = None,  
        overwrite: bool = True, 
        **kwargs
    ):
        """
        Upload a file to R2

        Utilize boto3
        """
        if not overwrite and await source.async_exists(): raise FileExistsError(f"{source} already exists and overwrite is False")
        filename = filename or self.name
        # s3t = self._accessor.s3t()
        self._accessor.s3t.upload(
            source.as_posix(),
            self._bucket,
            self.get_path_key(filename)
        )
        # self._accessor.s3t.shutdown()
        #await to_thread(
        #    self._accessor.boto.upload_file, Bucket = self._bucket, Key = self.get_path_key(filename), Filename = dest.as_posix()
        #)
        return self.parent.joinpath(filename)

    def batch_upload_files(
        self, 
        files: Optional[List['FileLike']] = None,
        glob_path: Optional[str] = None,
        overwrite: bool = False,
        skip_existing: bool = True,
        callbacks: Optional[List[Any]] = None,
        **kwargs
    ):
        """
        Handles batch uploading of files

        https://stackoverflow.com/questions/56639630/how-can-i-increase-my-aws-s3-upload-speed-when-using-boto3
        """
        assert files or glob_path, "Must provide either files or glob_path"
        if glob_path: files = list(self.glob(glob_path))
        results = []
        # s3t = self._accessor.s3t()
        for file in files:
            if not overwrite and skip_existing and file.exists(): continue
            self._accessor.s3t.upload(
                file.as_posix(),
                self._bucket,
                self.get_path_key(file.name),
                subscribers = callbacks
            )
            results.append(self.parent.joinpath(file.name))
        # self._accessor.s3t.shutdown()
        return results
    
    def batch_download_files(
        self,
        glob_path: str,
        output_dir: 'FileLike',
        overwrite: bool = False,
        skip_existing: bool = True,
        callbacks: Optional[List[Any]] = None,
        **kwargs
    ):
        """
        Handles batch downloading of files
        """
        files = list(self.glob(glob_path))
        results = []
        # s3t = self._accessor.s3t()
        for file in files:
            if not overwrite and skip_existing and file.exists(): continue
            output_file = output_dir.joinpath(file.name)
            self._accessor.s3t.download(
                self._bucket,
                self.get_path_key(file.name),
                output_file.as_posix(),
                subscribers = callbacks
            )
            results.append(output_file)
        self._accessor.s3t.shutdown()
        return results


    def unlink(self, missing_ok: bool = False):
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        try: self.filesys.rm_file(self._cloudpath)
        except FileNotFoundError:
            if not missing_ok: raise

    async def async_unlink(self, missing_ok: bool = False):
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        try: await self.afilesys._rm_file(self._cloudpath)
        except FileNotFoundError:
            if not missing_ok: raise

    def rm_file(self, missing_ok: bool = True):
        """
        Remove this file.
        If the path is a directory, use rmdir() instead.
        """
        try:
            self.filesys.rm_file(self._cloudpath)
            return True
        except Exception as e:
            if missing_ok: return False
            raise e from e

    async def async_rm_file(self, missing_ok: bool = True):
        """
        Remove this file.
        If the path is a directory, use rmdir() instead.
        """
        try:
            await self.afilesys._rm_file(self._cloudpath)
            return True
        except Exception as e:
            if missing_ok: return False
            raise e from e


class FileR2PosixPath(PosixPath, FileR2Path, PureFileR2PosixPath):
    __slots__ = ()


class FileR2WindowsPath(WindowsPath, FileR2Path, PureFileR2WindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("FileR2Path.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("FileR2Path.async_is_mount() is unsupported on this system")

register_pathlike(
    [
        FileR2PurePath, FileR2Path, PureFileR2PosixPath, FileR2WindowsPath, FileR2PosixPath, PureFileR2WindowsPath
    ]
)

R2FileSystem = R2_CloudFileSystem


__all__ = (
    'FileR2PurePath',
    'FileR2Path',
    'PureFileR2PosixPath',
    'FileR2WindowsPath',
    'FileR2PosixPath',
    'PureFileR2WindowsPath',
    'R2FileSystem'
)



from __future__ import annotations

import anyio
from fileio.lib.core import *
from fileio.lib.flavours import _pathz_windows_flavour, _pathz_posix_flavour
from fileio.lib.posix.static import _ASYNC_SYNTAX_MAPPING

from fileio.types.common import *
from fileio.utils.logs import logger

if TYPE_CHECKING:
    from fileio.lib.base import FilePath
    from fsspec.spec import AbstractBufferedFile
    from fsspec.asyn import AsyncFileSystem, AbstractAsyncStreamedFile
    from fileio.lib.posix.filesys import CloudFileSystemLike, AccessorLike
    from _typeshed import OpenBinaryMode, OpenTextMode, ReadableBuffer, WriteableBuffer

GIT_PREFIXES = ('gh://', 'git://', 'hf://')
URI_PREFIXES = ('gs://', 's3://', 'az://', 'minio://', 'mio://', 's3c://', 'r2://', 'wsbi://') + GIT_PREFIXES
_URI_SCHEMES = frozenset(('gs', 's3', 'az', 'minio', 'mio', 's3c', 'r2', 'wsbi', 'az', 'gh', 'git', 'hf'))
_URI_MAP_ROOT = {
    'gs://': '/gs/',
    's3://': '/s3/',
    'az://': '/azure/',
    'mio://': '/minio/',
    'minio://': '/minio/',
    's3c://': '/s3c/',
    'gh://': '/github/',
    'git://': '/git/',
    'hf://': '/huggingface/',
    'r2://': '/r2/',
    'wsbi://': '/wasabi/',
}
_PROVIDER_MAP = {
    'gs': 'GoogleCloudStorage',
    's3': 'AmazonS3',
    'az': 'Azure',
    'mio': 'MinIO',
    'minio': 'MinIO',
    's3c': 'S3Compatible',
    'gh': 'Github',
    'git': 'Git',
    'hf': 'HuggingFace',
    'r2': 'CloudFlare',
}

Paths = Union['FilePath', Path, str]


def rewrite_async_syntax(obj, provider: str = 's3'):
    """
    Basically - we're rewriting all the fsspec's async method
    from _method to async_method for syntax
    """
    _names = _ASYNC_SYNTAX_MAPPING[provider]
    for attr in dir(obj):
        if attr.startswith('_') and not attr.startswith('__'):
            attr_val = getattr(obj, attr)
            if iscoroutinefunction(attr_val) and _names.get(attr):
                setattr(obj, _names[attr], attr_val)
    return obj

@asynccontextmanager
async def get_cloud_handle(name: Paths, mode: FileMode = 'r', buffering: int = -1, encoding: str | None = ENCODING, errors: str | None = ERRORS, newline: str | None = SEP) -> AsyncContextManager[Handle]:
    file: AsyncFile
    if 'b' in mode: file = await open_file(name, mode)
    else: file = await open_file(name, mode, encoding=encoding, errors=errors, newline=newline)
    yield file
    await file.aclose()


@asynccontextmanager
async def get_cloud_file(filelike: Paths) -> AsyncContextManager[Handle]:
    file: AsyncFile
    filelike = cast(IO[Union[str, bytes, os.PathLike, Any]], filelike)
    file = AsyncFile(filelike)
    yield file
    await file.aclose()

class AsyncFileS(AsyncFile):
    """
    Wraps the file object to provide a more intuitive interface
    """

    def __init__(self, fp: Union['AbstractAsyncStreamedFile', IO[AnyStr]], is_asyn: Optional[bool] = False) -> None:
        self._fp: Union['AbstractAsyncStreamedFile', Any] = fp
        self._is_asyn = is_asyn

    async def aclose(self) -> None:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.close):
                return await self._fp.close()
            return await anyio.to_thread.run_sync(self._fp.close)
        return await super().aclose()

    async def read(self, size: int = -1) -> AnyStr:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.read):
                return await self._fp.read(size)
            return await anyio.to_thread.run_sync(self._fp.read, size)
        return super().read(size)

    async def read1(self, size: int = -1) -> bytes:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.read):
                return await self._fp.read(size)
            return await anyio.to_thread.run_sync(self._fp.read, size)
        return await super().read1(size)

    async def readline(self) -> AnyStr:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.readline):
                return await self._fp.readline()
            return await anyio.to_thread.run_sync(self._fp.readline)
        return await super().readline()

    async def readlines(self) -> list[AnyStr]:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.readlines):
                return await self._fp.readlines()
            return await anyio.to_thread.run_sync(self._fp.readlines)
        return await super().readlines()
    
    async def readinto(self, b: WriteableBuffer) -> bytes:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.readinto):
                return await self._fp.readinto(b)
            return await anyio.to_thread.run_sync(self._fp.readinto, b)
        return await super().readinto(b)

    async def readinto1(self, b: WriteableBuffer) -> bytes:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.readinto1):
                return await self._fp.readinto1(b)
            return await anyio.to_thread.run_sync(self._fp.readinto1, b)
        return await super().readinto1(b)

    async def write(self, b: ReadableBuffer | str) -> int:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.write):
                return await self._fp.write(b)
            return await anyio.to_thread.run_sync(self._fp.write, b)
        return await super().write(b)

    async def writelines(self, lines: Iterable[ReadableBuffer] | Iterable[str]) -> None:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.writelines):
                return await self._fp.writelines(lines)
            return await anyio.to_thread.run_sync(self._fp.writelines, lines)
        return await super().writelines(lines)
    
    async def truncate(self, size: int | None = None) -> int:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.truncate):
                return await self._fp.truncate(size)
            return await anyio.to_thread.run_sync(self._fp.truncate, size)
        return await super().truncate(size)

    async def seek(self, offset: int, whence: int | None = os.SEEK_SET) -> int:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.seek):
                return await self._fp.seek(offset, whence)
            return await anyio.to_thread.run_sync(self._fp.seek, offset, whence)
        return await super().seek(offset, whence)

    async def tell(self) -> int:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.tell):
                return await self._fp.tell()
            return await anyio.to_thread.run_sync(self._fp.tell)
        return await super().tell()

    async def flush(self, *args) -> None:
        if self._is_asyn:
            if iscoroutinefunction(self._fp.flush):
                return await self._fp.flush(*args)
            return await anyio.to_thread.run_sync(self._fp.flush, *args)
        return await super().flush()



@asynccontextmanager
async def get_cloudfs_file(accessor: 'AccessorLike', path: Paths, mode: FileMode = 'rb', **kwargs) -> AsyncContextManager[Handle]:
    """
    Helper function to open a file from a filesystem
    """
    if 'wb' in mode and hasattr(accessor.async_filesys, 'open_async'):
        asyncfile: 'AbstractAsyncStreamedFile' = await accessor.async_filesys.open_async(path, mode, **kwargs)
        if not hasattr(asyncfile, '_closed'):
            setattr(asyncfile, '_closed', False)
        # logger.warning(f'asyncfile: {asyncfile} - {type(asyncfile)} - {asyncfile.closed}')
        filelike = cast(IO[Union[str, bytes, os.PathLike, Any]], asyncfile)
        # file = AsyncFileS(filelike, is_asyn = True)            
    else:
        syncfile: 'AbstractBufferedFile' = accessor.open(path, mode, **kwargs)
        filelike = cast(IO[Union[str, bytes, os.PathLike, Any]], syncfile)
    
    file = AsyncFileS(filelike, is_asyn = True)
    try:
        yield file
    finally:
        await file.aclose()

    # if 'rb' in mode and hasattr(accessor.async_filesys, 'open_async'):
    # if hasattr(accessor.async_filesys, 'open_async'):
    #     try:
    #         asyncfile: 'AbstractAsyncStreamedFile' = await accessor.async_filesys.open_async(path, mode, **kwargs)
    #         filelike = cast(IO[Union[str, bytes, os.PathLike, Any]], asyncfile)
    #         file = AsyncFileS(filelike, is_asyn = True)            
    #         print('YIELDING FILE', type(file), file)
    #         yield file

    #     finally:
    #         print('CLOSING FILE')
    #         await file.close()
    #     file: 'AbstractAsyncStreamedFile' = await accessor.async_filesys.open_async(path, mode, **kwargs)
    #     yield file
    #     await file.close()
    # else:
    #     syncfile: 'AbstractBufferedFile' = accessor.open(path, mode, **kwargs)
    #     filelike = cast(IO[Union[str, bytes, os.PathLike, Any]], syncfile)
    #     file = AsyncFile(filelike)
    #     yield file
    #     await file.aclose()

    # file: AsyncFile
    # file = AsyncFile(filesys)
    # yield file
    # if hasattr(file, 'aclose'):
    #     await file.aclose()
    # else:
    #     await file.close()



# from __future__ import absolute_import

import os
# from . import types
from .lib.base import pathlib
from .types import (
    PathLike,
    FileInfo,
    PreparedFile,
    ParsedFile,
)

from .lib.types import (
    File,
    FileLike,
    FileType,
    FileListType,
    FileSysLike,
)

# from .types import (
#     File,
#     FileLike,
#     FileType,
#     FileListType,
#     PathLike,
#     FileSysLike,
#     FileInfo,
#     PreparedFile,
#     ParsedFile,
# )
from .utils import settings

__all__ = [
    'pathlib',
    'File',
    'FileLike',
    'FileType',
    'FileListType',
    'PathLike',
    'FileSysLike',
    'FileInfo',
    'PreparedFile',
    'ParsedFile',
    'settings',
]
# from fileio.core.imports import pathlib
# from fileio.core.generic import *

# from fileio.core import flavours
# from fileio.core import base
# from fileio.core import generic

# from fileio import providers

# from fileio.utils import settings
# from fileio.core.types import File, FileLike, FileType, FileListType

# from fileio.types.etc import ParsedFile, PreparedFile

# from fileio.converters.base import BaseConverter
# PreparedFile.update_forward_refs()
# ParsedFile.update_forward_refs()


import os
import warnings
from .base import *
from .utils import _log

with contextlib.suppress(ImportError):
    import s3fs
    from s3fs.errors import translate_boto_error
    from botocore.exceptions import ClientError
    from s3fs.core import _error_wrapper, sync_wrapper


if TYPE_CHECKING:
    from s3transfer.manager import TransferManager
    from aiobotocore.session import AioSession

class S3FileSystem(s3fs.S3FileSystem):
    """
    Access S3 as if it were a file system.

    This exposes a filesystem-like API (ls, cp, open, etc.) on top of S3
    storage.

    Provide credentials either explicitly (``key=``, ``secret=``) or depend
    on boto's credential methods. See botocore documentation for more
    information. If no credentials are available, use ``anon=True``.

    Parameters
    ----------
    anon : bool (False)
        Whether to use anonymous connection (public buckets only). If False,
        uses the key/secret given, or boto's credential resolver (client_kwargs,
        environment, variables, config files, EC2 IAM server, in that order)
    endpoint_url : string (None)
        Use this endpoint_url, if specified. Needed for connecting to non-AWS
        S3 buckets. Takes precedence over `endpoint_url` in client_kwargs.
    key : string (None)
        If not anonymous, use this access key ID, if specified. Takes precedence
        over `aws_access_key_id` in client_kwargs.
    secret : string (None)
        If not anonymous, use this secret access key, if specified. Takes
        precedence over `aws_secret_access_key` in client_kwargs.
    token : string (None)
        If not anonymous, use this security token, if specified
    use_ssl : bool (True)
        Whether to use SSL in connections to S3; may be faster without, but
        insecure. If ``use_ssl`` is also set in ``client_kwargs``,
        the value set in ``client_kwargs`` will take priority.
    s3_additional_kwargs : dict of parameters that are used when calling s3 api
        methods. Typically used for things like "ServerSideEncryption".
    client_kwargs : dict of parameters for the botocore client
    requester_pays : bool (False)
        If RequesterPays buckets are supported.
    default_block_size: int (None)
        If given, the default block size value used for ``open()``, if no
        specific value is given at all time. The built-in default is 5MB.
    default_fill_cache : Bool (True)
        Whether to use cache filling with open by default. Refer to
        ``S3File.open``.
    default_cache_type : string ("readahead")
        If given, the default cache_type value used for ``open()``. Set to "none"
        if no caching is desired. See fsspec's documentation for other available
        cache_type values. Default cache_type is "readahead".
    version_aware : bool (False)
        Whether to support bucket versioning.  If enable this will require the
        user to have the necessary IAM permissions for dealing with versioned
        objects. Note that in the event that you only need to work with the
        latest version of objects in a versioned bucket, and do not need the
        VersionId for those objects, you should set ``version_aware`` to False
        for performance reasons. When set to True, filesystem instances will
        use the S3 ListObjectVersions API call to list directory contents,
        which requires listing all historical object versions.
    cache_regions : bool (False)
        Whether to cache bucket regions or not. Whenever a new bucket is used,
        it will first find out which region it belongs and then use the client
        for that region.
    asynchronous :  bool (False)
        Whether this instance is to be used from inside coroutines.
    config_kwargs : dict of parameters passed to ``botocore.client.Config``
    kwargs : other parameters for core session.
    session : aiobotocore AioSession object to be used for all connections.
         This session will be used inplace of creating a new session inside S3FileSystem.
         For example: aiobotocore.session.AioSession(profile='test_user')

    The following parameters are passed on to fsspec:

    skip_instance_cache: to control reuse of instances
    use_listings_cache, listings_expiry_time, max_paths: to control reuse of directory listings

    Examples
    --------
    >>> s3 = S3FileSystem(anon=False)  # doctest: +SKIP
    >>> s3.ls('my-bucket/')  # doctest: +SKIP
    ['my-file.txt']

    >>> with s3.open('my-bucket/my-file.txt', mode='rb') as f:  # doctest: +SKIP
    ...     print(f.read())  # doctest: +SKIP
    b'Hello, world!'
    """

    root_marker = ""
    connect_timeout = 5
    retries = 5
    read_timeout = 15
    default_block_size = 5 * 2**20
    protocol = ["s3", "s3a"]
    _extra_tokenize_attributes = ("default_block_size",)

    s3tm: Optional['TransferManager'] = None

    def __init_dircache(self, **storage_options):
        """
        Initialize the directory cache
        """
        # self.dircache = DirCache(**storage_options)


    def __init_fs(self, *args, **storage_options):
        """Create and configure file-system instance

        Instances may be cachable, so if similar enough arguments are seen
        a new instance is not required. The token attribute exists to allow
        implementations to cache instances if they wish.

        A reasonable default should be provided if there are no arguments.

        Subclasses should call this method.

        Parameters
        ----------
        use_listings_cache, listings_expiry_time, max_paths:
            passed to ``DirCache``, if the implementation supports
            directory listing caching. Pass use_listings_cache=False
            to disable such caching.
        skip_instance_cache: bool
            If this is a cachable implementation, pass True here to force
            creating a new instance even if a matching instance exists, and prevent
            storing this instance.
        asynchronous: bool
        loop: asyncio-compatible IOLoop or None
        """
        # reusing instance, don't change
        if self._cached: return
        self._cached = True
        self._intrans = False
        self._transaction = None
        self._invalidated_caches_in_transaction = []
        self.__init_dircache(**storage_options)
        # self.dircache = DirCache(**storage_options)

        if storage_options.pop("add_docs", None):
            warnings.warn("add_docs is no longer supported.", FutureWarning)

        if storage_options.pop("add_aliases", None):
            warnings.warn("add_aliases has been removed.", FutureWarning)
        
        # This is set in _Cached
        self._fs_token_ = None

    def __init_afs(self, *args, asynchronous: Optional[bool] = False, loop: Optional[asyncio.AbstractEventLoop] = None, batch_size: Optional[int] = None, **kwargs):
        """
        The asynchronous version of __init_fs
        """
        self.asynchronous = asynchronous
        self._pid = os.getpid()
        self._loop = None if asynchronous else loop or get_loop()
        self.batch_size = batch_size
        self.__init_fs(*args, **kwargs)


    def __init__(
        self,
        anon: Optional[bool] = False,
        endpoint_url: Optional[str] = None,
        key: Optional[str] = None,
        secret: Optional[str] = None,
        token: Optional[str] = None,
        use_ssl: Optional[bool] = True,
        client_kwargs: Optional[Dict[str, Any]] = None,
        requester_pays: Optional[bool] = False,
        default_block_size: Optional[int] =None,
        default_fill_cache: Optional[bool] = True,
        default_cache_type: Optional[str] = "readahead",
        version_aware: Optional[bool] = False,
        config_kwargs: Optional[Dict[str, Any]] = None,
        s3_additional_kwargs: Optional[Dict[str, Any]]= None,
        session: Optional['AioSession'] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        cache_regions: Optional[bool] = False,
        asynchronous: Optional[bool] = False,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        **kwargs,
    ):
        if key and username:
            raise KeyError("Supply either key or username, not both")
        if secret and password:
            raise KeyError("Supply secret or password, not both")
        if username:
            key = username
        if password:
            secret = password

        self.endpoint_url = endpoint_url

        self.anon = anon
        self.key = key
        self.secret = secret
        self.token = token
        self.kwargs = kwargs
        super_kwargs = {
            k: kwargs.pop(k)
            for k in ["use_listings_cache", "listings_expiry_time", "max_paths"]
            if k in kwargs
        }  # passed to fsspec superclass
        self.__init_afs(loop = loop, asynchronous = asynchronous, **super_kwargs)
        # super().__init__(loop=loop, asynchronous=asynchronous, **super_kwargs)

        self.default_block_size = default_block_size or self.default_block_size
        self.default_fill_cache = default_fill_cache
        self.default_cache_type = default_cache_type
        self.version_aware = version_aware
        self.client_kwargs = client_kwargs or {}
        self.config_kwargs = config_kwargs or {}
        self.req_kw = {"RequestPayer": "requester"} if requester_pays else {}
        self.s3_additional_kwargs = s3_additional_kwargs or {}
        self.use_ssl = use_ssl
        self.cache_regions = cache_regions
        self._s3 = None
        self.session = session


    def _get_block_size(
        self,
        mode: str = "rb",
        block_size: Optional[int] = None,
    ) -> int:
        """
        Get the block size for reading or writing a file
        """
        if block_size is None:  return self.default_r2_write_block_size if 'w' in mode else self.default_r2_block_size
        return max(block_size, (self.default_r2_write_block_size if 'w' in mode else self.default_r2_block_size))


    def _open(
        self,
        path: str,
        mode: str = "rb",
        block_size: Optional[int] = None,
        acl: Optional[str] = "",
        version_id: Optional[str] = None,
        fill_cache: Optional[bool] = None,
        cache_type: Optional[str] = None,
        autocommit: Optional[bool] = True,
        requester_pays: Optional[bool] = None,
        cache_options: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> 'R2File':
        """Open a file for reading or writing

        Parameters
        ----------
        path: string
            Path of file on S3
        mode: string
            One of 'r', 'w', 'a', 'rb', 'wb', or 'ab'. These have the same meaning
            as they do for the built-in `open` function.
        block_size: int
            Size of data-node blocks if reading
        fill_cache: bool
            If seeking to new a part of the file beyond the current buffer,
            with this True, the buffer will be filled between the sections to
            best support random access. When reading only a few specific chunks
            out of a file, performance may be better if False.
        acl: str
            Canned ACL to set when writing
        version_id : str
            Explicit version of the object to open.  This requires that the s3
            filesystem is version aware and bucket versioning is enabled on the
            relevant bucket.
        encoding : str
            The encoding to use if opening the file in text mode. The platform's
            default text encoding is used if not given.
        cache_type : str
            See fsspec's documentation for available cache_type values. Set to "none"
            if no caching is desired. If None, defaults to ``self.default_cache_type``.
        requester_pays : bool (optional)
            If RequesterPays buckets are supported.  If None, defaults to the
            value used when creating the S3FileSystem (which defaults to False.)
        kwargs: dict-like
            Additional parameters used for s3 methods.  Typically used for
            ServerSideEncryption.
        """

        block_size = self._get_block_size(mode, block_size)
        if fill_cache is None: fill_cache = self.default_fill_cache
        if requester_pays is None: requester_pays = bool(self.req_kw)
        acl = acl or self.s3_additional_kwargs.get("ACL", "")
        kw = self.s3_additional_kwargs.copy()
        kw.update(kwargs)
        if not self.version_aware and version_id:
            raise ValueError(
                "version_id cannot be specified if the filesystem "
                "is not version aware"
            )

        if cache_type is None: cache_type = self.default_cache_type
        # _log(f's3tm: {self.s3tm}')
        return R2File(
            self,
            path,
            mode,
            block_size=block_size,
            acl=acl,
            version_id=version_id,
            fill_cache=fill_cache,
            s3_additional_kwargs=kw,
            cache_type=cache_type,
            autocommit=autocommit,
            requester_pays=requester_pays,
            cache_options=cache_options,
        )

    
    async def _rm_file(self, path, **kwargs):
        bucket, key, _ = self.split_path(path)
        delete_keys = {
            "Objects": [{"Key": key}],
            "Quiet": True,
        }
        self.invalidate_cache(self._parent(path))
        # logger.info(f'Deleting {path}: {delete_keys}')
        try:
            # await self._call_s3("delete_object", Bucket=bucket, Key=key)
            await self._call_s3("delete_objects", Bucket=bucket, Delete = delete_keys)
        except ClientError as e:
            raise translate_boto_error(e) from e
        
    
    async def _call_s3(self, method, *akwarglist, **kwargs):
        await self.set_session()
        s3 = await self.get_s3(kwargs.get("Bucket"))
        method = getattr(s3, method)
        kw2 = kwargs.copy()
        kw2.pop("Body", None)
        _log(f"CALL: {method.__name__} - {akwarglist} - {kw2}")
        additional_kwargs = self._get_s3_method_kwargs(method, *akwarglist, **kwargs)
        return await _error_wrapper(
            method, kwargs=additional_kwargs, retries=self.retries
        )
    
    call_s3 = sync_wrapper(_call_s3)

    async def open_async(self, path, mode="rb", block_size: Optional[int] = None, **kwargs):
        if "b" not in mode: # or kwargs.get("compression"):
            raise ValueError
        _ = kwargs.pop("compression", None)

        block_size = self._get_block_size(mode, block_size)
        return R2AsyncStreamedFile(self, path, mode, block_size = block_size, **kwargs)



from __future__ import absolute_import

from fileio.types.common import *
from fileio.types.classprops import lazyproperty
# from fileio.types.core import File, FileLike, FileType, FileListType, PathLike, FileSysLike
from fileio.types.etc import FileInfo, PreparedFile, ParsedFile

